# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import dataclasses
import typing as t

from ._messages import DAPObjectMeta


@dataclasses.dataclass()
class Breakpoint(
    metaclass=DAPObjectMeta,
    dap={
        "id": "id",
        "verified": "verified",
        "message": "message",
        "source": "source",
        "line": "line",
        "column": "column",
        "endLine": "end_line",
        "endColumn": "end_column",
        "instructionReference": "instruction_reference",
        "offset": "offset",
    },
):
    """Breakpoint information.

    Information about a breakpoint created in SetBreakpointsRequest,
    SetFunctionBreakpointsRequest, SetInstructionBreakspointsRequest, or
    SetDataBreakpointsRequest.

    Args:
        id: Optional identifier for the breakpoint.
        verified: The breakpoint could be set.
        message: Option explanation on the state of the breakpoint.
        source: Source where the breakpoint is located.
        line: The start line of the actual range covered by the breakpoint.
        column: The start column of the actual range covered by the breakpoint.
        end_line: End line of the actual range covered by the breakpoint.
        end_column: End column of the actual range covered by the breakpoint.
            If no end_line is specified, then the column is assumed to be in
            the start_line.
        instruction_reference: Optional memory reference to where the
            breakpoint is set.
        offset: Optional offset from the instruction reference.
    """

    id: t.Optional[int] = None
    verified: bool = False
    message: t.Optional[str] = None
    source: t.Optional[Source] = None
    line: t.Optional[int] = None
    column: t.Optional[int] = None
    end_line: t.Optional[int] = None
    end_column: t.Optional[int] = None
    instruction_reference: t.Optional[str] = None
    offset: t.Optional[int] = None


@dataclasses.dataclass()
class Capabilities(
    metaclass=DAPObjectMeta,
    dap={
        "supportsConfigurationDoneRequest": "supports_configuration_done_request",
        "supportsFunctionBreakpoints": "supports_function_breakpoints",
        "supportsConditionalBreakpoints": "supports_conditional_breakpoints",
        "supportsHitConditionalBreakpoints": "supports_hit_conditional_breakpoints",
        "supportsEvaluateForHovers": "supports_evaluate_for_hovers",
        "supportsStepBack": "supports_step_back",
        "supportsSetVariable": "supports_set_variable",
        "supportsRestartFrame": "supports_restart_frame",
        "supportsGotoTargetsRequest": "supports_goto_targets_request",
        "supportsStepInTargetsRequest": "supports_step_in_targets_request",
        "supportsCompletionsRequest": "supports_completions_request",
        "supportTerminateDebuggee": "supports_terminate_debuggee",
        "supportSuspendDebuggee": "supports_suspend_debuggee",
        "supportsTerminateRequest": "supports_terminate_request",
        "supportsClipboardContext": "supports_clipboard_context",
    },
):
    """Capabilities of a debug adapter.

    Information about the capabilities of a debug adapter.

    Args:
        supports_configuration_done_request: The debug adapter supports the
            ConfigurationDoneRequest.
        supports_function_breakpoints: The debug adapter supports function
            breakpoints.
        supports_conditional_breakpoints: The debug adapter supports conditional
            breakpoints.
        supports_hit_conditional_breakpoints: The debug adapter supports
            breakpoints that break execution after a specified number of hits.
        supports_evaluate_for_hovers: THe debug adapter supports evaluate
            request for data hovers.
        supports_step_back: The debug adapter supports stepping back via the
            StepBackRequest and ReverseContinueRequest.
        supports_set_variable: The debug adapter supports setting a variable to
            a value.
        supports_restart_frame: The debug adapter supports restarting a frame.
        supports_goto_targets_request: The debug adapter supports the
            GotoTargetsRequest.
        supports_step_in_targets_request: The debug adapter supports the
            StepInTargetsRequest.
        supports_completions_request: The debug adapter supports the
            CompletionsRequest.
        supports_terminate_debuggee: The debug adapter supports the
            terminate_debuggee` attribute on the DisconnectRequest.
        supports_suspend_debuggee: The debug adapter supports the
            suspend_debuggee attribute on the DisconnectRequest.
        supports_terminate_request: The debug adapter supports the
            TerminateRequest.
        supports_clipboard_context: The debug adapter supports the clipboard
            context value in the evaluate request.
    """

    supports_configuration_done_request: bool = False
    supports_function_breakpoints: bool = False
    supports_conditional_breakpoints: bool = False
    supports_hit_conditional_breakpoints: bool = False
    supports_evaluate_for_hovers: bool = False
    # exception_breakpoint_filters
    supports_step_back: bool = False
    supports_set_variable: bool = False
    supports_restart_frame: bool = False
    supports_goto_targets_request: bool = False
    supports_step_in_targets_request: bool = False
    supports_completions_request: bool = False
    # completion_trigger_characters
    # supports_modules_request
    # additional_module_columns
    # supported_checksum_algorithms
    # supports_restart_request
    # supports_exception_options
    # supports_value_formatting_options
    # supports_exception_info_request
    supports_terminate_debuggee: bool = False
    supports_suspend_debuggee: bool = False
    # supports_delayed_stack_trace_loading
    # supports_loaded_sources_request
    # supports_log_points
    # supports_terminate_threads_request
    # supports_set_expression
    supports_terminate_request: bool = False
    # supports_data_breakpoints
    # supports_read_memory_request
    # supports_write_memory_request
    # supports_disassemble_request
    # supports_cancel_request
    # supports_breakpoint_locations_request
    supports_clipboard_context: bool = False
    # supports_stepping_granularity
    # supports_instruction_breakpoints
    # supports_exception_filter_options
    # supports_single_thread_execution_requests


@dataclasses.dataclass()
class Checksum(metaclass=DAPObjectMeta, dap={"algorithm": "algorithm", "checksum": "checksum"}):
    """The checksum of an item.

    Describes the checksum of an item. The known algorithms are "MD5", "SHA1",
    "SHA256", and "timestamp".

    Args:
        algorithm: The algorithm used to calculate this checksum.
        checksum: The value of the checksum, encoded as a hexadecimal value.
    """

    algorithm: t.Literal["MD5", "SHA1", "SHA256", "timestamp"]
    checksum: str


@dataclasses.dataclass()
class ExceptionFilterOptions(metaclass=DAPObjectMeta, dap={"filterId": "filter_id", "condition": "condition"}):
    """Exception filter options.

    Used to specify an exception filter together with a condition for the
    SetExceptionBreakpoints request.

    Args:
        filter_id: The ID of the exception filter.
        condition: Optional condition for the filter.
    """

    filter_id: str
    condition: t.Optional[str] = None


@dataclasses.dataclass()
class ExceptionOptions(metaclass=DAPObjectMeta, dap={"path": "path", "breakMode": "break_mode"}):
    """Configuration options to a set of exceptions.

    Assigns configuration options to a set of exceptions.

    Args:
        path: A path that selects a single or multiple exceptions in a tree.
            By convention the first segment of the path is a category that is
            used to group exceptions in the UI.
        break_mode: Condition when a thrown exception should result in a break.
    """

    path: t.List[ExceptionPathSegment]
    break_mode: t.Literal["never", "always", "unhandled", "userUnhandled"]


@dataclasses.dataclass()
class ExceptionPathSegment(metaclass=DAPObjectMeta, dap={"negate": "negate", "names": "names"}):
    """Represents a segment in a path.

    Represents a segment in a path that is used to match leafs or nodes in a
    tree of exceptions.

    Args:
        negate: Controls the matching behaviour of the names.
        names: The values to match (or not match if negate=True).
    """

    negate: bool = False
    names: t.List[str] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class Message(
    metaclass=DAPObjectMeta,
    dap={
        "id": "id",
        "format": "format",
        "variables": "variables",
        "sendTelemetry": "send_telemetry",
        "showUser": "show_user",
        "url": "url",
        "urlLabel": "url_label",
    },
):
    """A structured message object.

    A structured message object used to return errors from requests.
    """

    id: int
    format: str
    variables: t.Dict[str, str] = dataclasses.field(default_factory=dict)
    send_telemetry: bool = False
    show_user: bool = False
    url: t.Optional[str] = None
    url_label: t.Optional[str] = None


@dataclasses.dataclass()
class Scope(
    metaclass=DAPObjectMeta,
    dap={
        "name": "name",
        "presentationHint": "presentation_hint",
        "variablesReference": "variables_reference",
        "namedVariables": "named_variables",
        "indexedVariables": "indexed_variables",
        "expensive": "expensive",
        "source": "source",
        "line": "line",
        "column": "column",
        "endLine": "end_line",
        "endColumn": "end_column",
    },
):
    """Named container for variables.

    A scope is a ned container for variables.

    Args:
        name: The name of the scope to be shown in the UI.
        variables_reference: The id used to retrieve the variables of this
            scope.
        presentation_hint: How to present this scope in the UI.
        named_variables: Number of named variables in the scope.
        expensive: If True, the number of variables in this scope is large or
            expensive to retrieve.
        source: Optional source for this scope.
        line: Optional start line of the range covered by this scope.
        column: Optional start column of the range covered by this scope.
        end_line: Optional end line of the range covered by this scope.
        end_column: Optional end column of the range covered by this scope.
    """

    name: str
    variables_reference: int
    presentation_hint: t.Optional[t.Union[str, t.Literal["arguments", "locals", "registers"]]] = None
    named_variables: t.Optional[int] = None
    indexed_variables: t.Optional[int] = None
    expensive: bool = False
    source: t.Optional[Source] = None
    line: t.Optional[int] = None
    column: t.Optional[int] = None
    end_line: t.Optional[int] = None
    end_column: t.Optional[int] = None


@dataclasses.dataclass()
class Source(
    metaclass=DAPObjectMeta,
    dap={
        "name": "name",
        "path": "path",
        "sourceReference": "source_reference",
        "presentationHint": "presentation_hint",
        "origin": "origin",
        "sources": "sources",
        "adapterData": "adapter_data",
        "checksums": "checksums",
    },
):
    """Descriptor for source code.

    A source is used to describe source code. It is returned from the debug
    adapter as part of a StackFrame and it is used by clients when specifying
    breakpoints.

    Args:
        name: The short name of the source.
        path: The path of the source to be shown in the UI. It is used to
            locate the source if source_reference is greater than 0.
        source_reference: If greater than 0, the contents of the source must be
            retrieved through the SourceRequest. The id is only valid for a
            session.
        presentation_hint: How to present the source in the UI. A value of
            deemphasize can be used to indicate the source is not available or
            that it is skipped on stepping.
        origin: Origin of this source.
        sources: List of sources that are related to this source.
        adapter_data: Optional opaque data to associate with the source. The
            client does not interpret this data.
        checksums: Checksum associated with this file.
    """

    name: t.Optional[str] = None
    path: t.Optional[str] = None
    source_reference: int = 0
    presentation_hint: t.Literal["normal", "emphasize", "deemphasize"] = "normal"
    origin: t.Optional[str] = None
    sources: t.List[Source] = dataclasses.field(default_factory=list)
    adapter_data: t.Any = None
    checksums: t.List[Checksum] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class SourceBreakpoint(
    metaclass=DAPObjectMeta,
    dap={
        "line": "line",
        "column": "column",
        "condition": "condition",
        "hitCondition": "hit_condition",
        "logMessage": "log_message",
    },
):
    """Properties of a breakpoint.

    The properties of a breakpoint or logpoint passed to the
    SetBreakpoinsRequest.

    Args:
        line: The source line of the breakpoint or logpoint.
        column: The optional source column of the breakpoint.
        condition: An optional expression for conditional breakpoints.
        hit_condition: An optional expression that controls how many hits of
            the brekapoint are ignored.
        log_message: Do not break but log the message on a break.
    """

    line: int
    column: t.Optional[int] = None
    condition: t.Optional[str] = None
    hit_condition: t.Optional[str] = None
    log_message: t.Optional[str] = None


@dataclasses.dataclass()
class StackFrame(
    metaclass=DAPObjectMeta,
    dap={
        "id": "id",
        "name": "name",
        "source": "source",
        "line": "line",
        "column": "column",
        "endLine": "end_line",
        "endColumn": "end_column",
        "canRestart": "can_restart",
        "instructionPointerReference": "instruction_pointer_reference",
        "moduleId": "module_id",
        "presentationHint": "presentation_hint",
    },
):
    """A stackframe.

    A stackframe contains the source location.

    Args:
        id: Id for the stack frame.
        name: The name of the stack frame.
        source: The optional source of the frame.
        line: The line within the file of the frame, will be ignored if source
            is None.
        column: The column within the file of the frame, will be ignored if
            source is None.
        end_line: Optional end line of the range covered by the stack frame.
        end_column: Optional end column of the range covered by the stack frame.
        can_restart: Indicates whether this frame can be restarted.
        instruction_pointer_reference: Optional memory reference for the
            current instruction pointer.
        module_id: Optional module associated with this frame.
        presentation_hint: How to present this frame in the UI.
    """

    id: int
    name: str
    source: t.Optional[Source] = None
    line: int = 0
    column: int = 0
    end_line: t.Optional[int] = None
    end_column: t.Optional[int] = None
    can_restart: bool = False
    instruction_pointer_reference: t.Optional[str] = None
    module_id: t.Optional[t.Union[int, str]] = None
    presentation_hint: t.Literal["normal", "label", "subtle"] = "normal"


@dataclasses.dataclass()
class StackFrameFormat(
    metaclass=DAPObjectMeta,
    dap={
        "parameters": "parameters",
        "parameterTypes": "parameter_types",
        "parameterNames": "parameter_names",
        "parameterValues": "parameter_values",
        "line": "line",
        "module": "module",
        "includeAll": "include_all",
    },
):
    """Formatting info for a stack frame.

    Provides formatting information for a stack frame.

    Args:
        parameters: Display parameters for the stack frame.
        parameter_types: Displays the types of parameters for the stack frame.
        parameter_names: Displays the names of parameters for the stack frame.
        parameter_values: Displays the values of parameters for the stack frame.
        line: Displays the line number of the stack frame.
        module: Displays the module of the stack frame.
        include_all: Includes all stack frames, including those the debug
            adapter might otherwise hide.
    """

    parameters: bool = False
    parameter_types: bool = False
    parameter_names: bool = False
    parameter_values: bool = False
    line: bool = False
    module: bool = False
    include_all: bool = False


@dataclasses.dataclass()
class Thread(metaclass=DAPObjectMeta, dap={"id": "id", "name": "name"}):
    """A thread.

    Represents a thread on the debuggee.

    Args:
        id: The unique identifier for the thread.
        name: The name of the thread.
    """

    id: int
    name: str


@dataclasses.dataclass()
class ValueFormat(metaclass=DAPObjectMeta, dap={"hex": "hex"}):
    """Provides formatting information for a value.

    Args:
        hex: Display the value in hex.
    """

    hex: bool = False


@dataclasses.dataclass()
class Variable(
    metaclass=DAPObjectMeta,
    dap={
        "name": "name",
        "value": "value",
        "type": "type",
        "presentationHint": "presentation_hint",
        "evaluateName": "evaluate_name",
        "variablesReference": "variables_reference",
        "namedVariables": "named_variables",
        "indexedVariables": "indexed_variables",
        "memoryReference": "memory_reference",
    },
):
    """A variable is a name/value pair.

    Represents a variable with a name and value as well as other metadata to
    display in the client GUI.

    Args:
        name: The variable's name.
        value: The variable's value.
        type: The type of the variable's value to display in the client.
        presentation_hint: Properties of a variable that can be used to
            determine how to render the variable in the client.
        evaluate_name: Optional evaluable name of this variable which can be
            passed to the EvaluateRequest to fetch the variable's name.
        variables_reference: If set to > 1, the variable is structured and
            contains indexed or named variables.
        named_variables: The number of named child variables.
        indexed_variables: The number of indexed child variables.
        memory_reference: Optional memory reference for the variable if the
            variable represents executable code, such as a function pointer.
    """

    name: str
    value: str
    type: t.Optional[str] = None
    presentation_hint: t.Optional[VariablePresentationHint] = None
    evaluate_name: t.Optional[str] = None
    variables_reference: int = 0
    named_variables: t.Optional[int] = None
    indexed_variables: t.Optional[int] = None
    memory_reference: t.Optional[str] = None


@dataclasses.dataclass
class VariablePresentationHint(
    metaclass=DAPObjectMeta,
    dap={
        "kind": "kind",
        "attributes": "attributes",
        "visibility": "visibility",
        "lazy": "lazy",
    },
):
    """Optional properties of a variable.

    Optional properties of a variable that can be used to determine how to
    render the variable in the UI.

    Args:
        kind: The kind of variable.
        attributes: Set of attributes represented as a list of strings.
        visbility: Visibility of the variable.
        lazy: The value can be retrieved through a specific request.
    """

    kind: t.Optional[
        t.Union[
            str,
            t.Literal[
                "property",
                "method",
                "class",
                "data",
                "event",
                "baseClass",
                "innerClass",
                "interface",
                "mostDerivedClass",
                "virtual",
                "dataBreakpoint",
            ],
        ]
    ] = None
    attributes: t.List[
        t.Union[
            str,
            t.Literal[
                "static",
                "constant",
                "readOnly",
                "rawString",
                "hasObjectId",
                "canHaveObjectId",
                "hasSideEffect",
                "hasDataBreakpoint",
            ],
        ]
    ] = dataclasses.field(default_factory=list)
    visibility: t.Optional[t.Union[str, t.Literal["public", "private", "protected", "internal"]]] = None
    lazy: bool = False



# Copyright (c) 2023 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

from ansible.playbook.block import Block
from ansible.playbook.play import Play
from ansible.playbook.task import Task

from ansibug._debuggee import AnsibleDebugger


def _split_task_path(task_path: str) -> tuple[str, int]:
    split = task_path.rsplit(":", 1)
    return split[0], int(split[1])


def register_block_breakpoints(
    debugger: AnsibleDebugger,
    blocks: list[Block],
) -> None:
    """Registers the block breakpoint locations.

    Registers all the valid breakpoints that can be found in the provided
    blocks with the debugger object. This is used to validate breakpoint
    locations as they are found.

    Args:
        debugger: The AnsibleDebugger object to register the breakpoints on.
        blocks: The blocks to scan for breakpoints.
    """
    # Need to register these blocks as valid breakpoints and update the client bps
    for block in blocks:
        block_path_and_line = block.get_path()
        if block_path_and_line:
            # If the path is set this is an explicit block and should be
            # marked as an invalid breakpoint section.
            block_path, block_line = _split_task_path(block_path_and_line)
            debugger.register_path_breakpoint(block_path, block_line, 0)

        task_list: list[Task] = block.block[:]
        task_list.extend(block.rescue)
        task_list.extend(block.always)

        for task in task_list:
            if isinstance(task, Block):
                # import_tasks will wrap the tasks in a standalone block.
                register_block_breakpoints(debugger, [task])

            else:
                # 2.19 changed how the play's flush_handlers and implicit
                # no-op tasks are setup. It no longer has the loader path and
                # play details so skip registering these if not set.
                raw_task_path = task.get_path()
                if not raw_task_path:
                    continue

                task_path, task_line = _split_task_path(raw_task_path)
                debugger.register_path_breakpoint(task_path, task_line, 1)


def register_play_breakpoints(
    debugger: AnsibleDebugger,
    plays: list[Play],
) -> None:
    """Registers the play breakpoint locations.

    Registers all the valid breakpoints that can be found in the provided plays
    with the debugger object. This is used to validate breakpoint locations as
    they are found.

    Args:
        debugger: The AnsibleDebugger object to register the breakpoints on.
        plays: The plays to scan for breakpoints.
    """
    for play in plays:
        # This is essentially doing what play.compile() does but without the
        # flush stages.
        play_path, play_line = _split_task_path(play.get_path())
        debugger.register_path_breakpoint(play_path, play_line, 1)

        play_blocks = play.compile() + play.handlers
        for r in play.roles:
            play_blocks += r.get_handler_blocks(play)

        register_block_breakpoints(debugger, play_blocks)



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import json

from ._messages import DAPEncoder, ProtocolMessage


class DebugAdapterConnection:
    def __init__(self) -> None:
        self._out_buffer = bytearray()
        self._in_buffer = bytearray()
        self.__seq_no_in = 1
        self.__seq_no_out = 1

    @property
    def _seq_no_in(self) -> int:
        no = self.__seq_no_in
        self.__seq_no_in += 1
        return no

    @property
    def _seq_no_out(self) -> int:
        no = self.__seq_no_out
        self.__seq_no_out += 1
        return no

    def data_to_send(
        self,
        n: int = -1,
    ) -> bytes:
        if n == -1:
            n = len(self._out_buffer)
        n = min(n, len(self._out_buffer))

        data = bytes(self._out_buffer[:n])
        self._out_buffer = self._out_buffer[n:]

        return data

    def receive_data(
        self,
        data: bytes | bytearray | memoryview,
    ) -> None:
        self._in_buffer += data

    def next_message(self) -> ProtocolMessage | None:
        header_idx = self._in_buffer.find(b"\r\n\r\n")
        if header_idx == -1:
            return None

        raw_headers = bytes(self._in_buffer[:header_idx])
        headers = dict(h.split(b": ", 2) for h in raw_headers.split(b"\r\n"))
        if b"Content-Length" not in headers:
            raise ValueError("Expected Content-Length header before message payload but none found")

        length = int(headers[b"Content-Length"])
        header_idx += 4

        if (len(self._in_buffer) - header_idx) < length:
            return None

        raw_msg = self._in_buffer[header_idx : header_idx + length]
        self._in_buffer = self._in_buffer[header_idx + length :]

        msg_data = json.loads(raw_msg.decode())
        msg = ProtocolMessage.unpack(msg_data)
        expected_seq_no_in = self._seq_no_in
        if expected_seq_no_in != msg.seq:
            raise ValueError(f"Expected seq {expected_seq_no_in} but received {msg.seq}")

        return msg

    def queue_msg(
        self,
        msg: ProtocolMessage,
    ) -> int:
        seq_no = self._seq_no_out
        msg.seq = seq_no
        data = msg.pack()
        serialized_msg = json.dumps(data, cls=DAPEncoder)

        self._out_buffer += f"Content-Length: {len(serialized_msg)}\r\n\r\n{serialized_msg}".encode()

        return seq_no



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
name: debug
type: aggregate
short_description: Ansible Debug Adapter Protocol Callback Plugin
description:
- The callback plugin used to handle the DAP socket connections.
- Callers should use the C(ansibug) module to invoke ansible-playbook with the
  required debug plugins enabled rather than do it directly.
author: Jordan Borean (@jborean93)
options:
  mode:
    description:
    - The socket mode to use.
    - C(connect) will connect to the addr requested and bound on the debug
      adapter server.
    - C(listen) will bind a new socket to the addr requested and wait for a
      debug adapter server to connect.
    type: str
    choices:
    - connect
    - listen
    env:
    - name: ANSIBUG_MODE
  debug_addr:
    description:
    - For C(mode=listen), the connection string the process will listen on.
    - For C(mode=connect), the connection string the process will connect to.
    - The details of these connection strings are internal to ansibug itself.
    default: localhost
    required: true
    type: str
    env:
    - name: ANSIBUG_DEBUG_ADDR
  tls_server_certfile:
    description:
    - The TLS server certificate used when C(mode=listen).
    - This is the path to a single file in PEM format containing the
      certificate, as well as any number of CA certificates needed to establish
      the certificate's authenticity.
    - This can also contain the PEM encoded certificate key, otherwise use
      C(tls_server_keyfile).
    type: path
    env:
    - name: ANSIBUG_TLS_SERVER_CERTFILE
  tls_server_keyfile:
    description:
    - The TLS server certificate key used when C(mode=listen).
    - This is the path to a single file in PEM format containing the
      certificate key.
    - If the key is encrypted use C(tls_server_key_password) to provide the
      password needed to decrypt the key.
    type: path
    env:
    - name: ANSIBUG_TLS_SERVER_KEYFILE
  tls_server_key_password:
    description:
    - The password needed to decrypt the key specified by
      C(tls_server_certfile) or C(tls_server_keyfile).
    type: str
    env:
    - name: ANSIBUG_TLS_SERVER_KEY_PASSWORD
  tls_client_ca:
    description:
    - The TLS client authentication CA verification bundles to use when
      C(mode=listen).
    - If set this will enforce client authentication meaning the client
      connecting to the listener must provide the certificate and key issued by
      a CA in the bundle provided.
    - If not set then no client authentication will be performed.
    type: path
    env:
    - name: ANSIBUG_TLS_CLIENT_CA
  use_tls:
    description:
    - Sets up a TLS protected stream when C(mode=listen).
    - Use C(tls_server_certfile), C(tls_server_keyfile), and
      C(tls_server_key_password) to specify the certificate and key used as the
      current identity.
    default: False
    type: bool
    env:
    - name: ANSIBUG_USE_TLS
  no_wait_for_config_done:
    description:
    - If true, will wait until the DA server has passed through the
      configurationDone request from the client that indicates all the initial
      breakpoint configuration has been sent through and the client is ready to
      run the code.
    - If false, will start the playbook immediately without waiting for the
      initial configuration details.
    type: bool
    default: false
    env:
    - name: ANSIBUG_NO_WAIT_FOR_CONFIG_DONE
  log_file:
    description:
    - The path used to store background logging events of the DAP thread.
    type: path
    env:
    - name: ANSIBUG_LOG_FILE
  log_level:
    description:
    - The log level to enable if C(log_file) is set.
    type: str
    default: info
    choices:
    - info
    - debug
    - warning
    - error
    env:
    - name: ANSIBUG_LOG_LEVEL
  log_format:
    description:
    - The log format to apply to C(log_file) if set
    type: str
    default: '%(asctime)s | %(name)s | %(filename)s:%(lineno)s %(funcName)s() %(message)s'
    env:
    - name: ANSIBUG_LOG_FORMAT
"""

import logging
import os
import ssl
import typing as t

from ansible.executor.stats import AggregateStats
from ansible.playbook import Playbook
from ansible.plugins.callback import CallbackBase
from ansible.utils.display import Display

from ansibug._debuggee import AnsibleDebugger
from ansibug._logging import configure_file_logging
from ansibug._tls import create_server_tls_context

from ..plugin_utils._breakpoints import register_play_breakpoints

log = logging.getLogger("ansibug.callback")

display = Display()


class CallbackModule(CallbackBase):
    CALLBACK_VERSION = 2.0
    CALLBACK_TYPE = "aggregate"
    CALLBACK_NAME = "debug"
    CALLBACK_NEEDS_ENABLED = True

    def __init__(
        self,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._debugger = AnsibleDebugger()

    def v2_playbook_on_start(
        self,
        playbook: Playbook,
    ) -> None:
        register_play_breakpoints(self._debugger, playbook.get_plays())

        log_file = self.get_option("log_file")
        if log_file:
            configure_file_logging(
                log_file,
                self.get_option("log_level"),
                self.get_option("log_format"),
            )

        mode = self.get_option("mode")
        debug_addr = self.get_option("debug_addr")

        ssl_context: ssl.SSLContext | None = None
        if self.get_option("use_tls") and mode == "listen":
            ssl_context = create_server_tls_context(
                certfile=self.get_option("tls_server_certfile"),
                keyfile=self.get_option("tls_server_keyfile"),
                password=self.get_option("tls_server_key_password"),
                ca_trust=self.get_option("tls_client_ca"),
            )

        log.info("Staring Ansible Debugger with %s on '%s'", mode, debug_addr)
        socket_addr = self._debugger.start(
            addr=debug_addr,
            mode=mode,
            ssl_context=ssl_context,
            playbook_file=getattr(playbook, "_file_name", None),
        )

        if mode == "listen":
            display.display(f"Ansibug listener has been configured for process PID {os.getpid()} on '{socket_addr}'")

        no_wait_for_config_done = self.get_option("no_wait_for_config_done")
        if not no_wait_for_config_done:
            log.info("Waiting for configuration done request to be received in Ansible")
            self._debugger.wait_for_config_done(timeout=None)

    def v2_playbook_on_stats(
        self,
        stats: AggregateStats,
    ) -> None:
        log.info("Shutting down Ansible Debugger")
        self._debugger.shutdown()



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import atexit
import json
import os
import pathlib
import subprocess
import sys
import typing as t

from ._logging import LogLevel


def exec_playbook_connect(
    playbook_args: list[str],
    addr: str,
    *,
    no_wait: bool = True,
    log_file: pathlib.Path | None = None,
    log_level: LogLevel = "info",
) -> None:
    """Launch a new debug ansible-playbook process in connect mode.

    Execs a new ansible-playbook process that will connect to the debug server
    address listed. This will use the os.exec method to start the new process
    to replace the existing one with the env vars needed for ansibug to work.

    If use_tls is set to True the client will wrap the socket connection in a
    TLS tunnel to encrypt the data exchanged. By default it will attempt to
    verify the server's identity through its certificate. The kwarg tls_cert_ca
    can be set to a boolean that turns the verification process on or off. It
    can also be set to a string value that is the path to a CA file or CA
    directory of PEM files to use as the CA trust store(s).

    Args:
        playbook_args: Arguments to invoke ansible-playbook with.
        addr: The address to connect to.
        no_wait: Do not wait until the client has communicated with the
            ansible-playbook process and sent the configurationDone request
            before starting the playbook.
        log_file: Set ansibug debuggee logger to log to the absolute path of
            this file if set.
        log_level: Set ansibug debuggee logger filter to use this level when
            logging. This only applies if log_file is also set.
    """
    _exec_playbook(
        playbook_args=playbook_args,
        mode="connect",
        mode_env={},
        addr=addr,
        no_wait=no_wait,
        use_tls=False,
        log_file=log_file,
        log_level=log_level,
    )


def exec_playbook_listen(
    playbook_args: list[str],
    addr: str,
    *,
    no_wait: bool = True,
    use_tls: bool = False,
    tls_cert: pathlib.Path | None = None,
    tls_key: pathlib.Path | None = None,
    tls_password: str | None = None,
    tls_client_ca: pathlib.Path | None = None,
    log_file: pathlib.Path | None = None,
    log_level: LogLevel = "info",
) -> None:
    """Launch a new debug ansible-playbook process in listen mode.

    Execs a new ansible-playbook process that will listen for the debug server
    to connect to it. If no addr is specified, the port is provided by the OS.
    This will use the os.exec method to start the new process to replace the
    existing one with the env vars needed for ansibug to work.

    If use_tls is set to True the ansible-playbook process will wrap the socket
    connection in a TLS tunnel to encrypt the data exchanged. It will use the
    existing OS CA settings when attempting to create the TLS server endpoint
    but the tls_cert, tls_key, and tls_password kwargs can be used to provide
    a custom certificate and key. The tls_cert file can contain both the
    certificate and key, if they are in separate files use the tls_key kwarg
    to specify the key path.

    Args:
        playbook_args: Arguments to invoke ansible-playbook with.
        addr: The address to listen on.
        no_wait: Do not wait until the client has communicated with the
            ansible-playbook process and sent the configurationDone request
            before starting the playbook.
        use_tls: New process will wrap the socket connection through a TLS
            tunnel.
        tls_cert: The path to the TLS certificate PEM encoded file to use for
            TLS.
        tls_key: The path to the TLS key PEM encoded file to use for TLS.
        tls_password: The password for the tls_key if it is encrypted.
        tls_client_ca: Optional CA bundle that will enforce TLS client
            authentication with a cert signed by a CA in the bundle.
        log_file: Set ansibug debuggee logger to log to the absolute path of
            this file if set.
        log_level: Set ansibug debuggee logger filter to use this level when
            logging. This only applies if log_file is also set.
    """
    mode_env = {}

    if tls_cert:
        mode_env["ANSIBUG_TLS_SERVER_CERTFILE"] = str(tls_cert.absolute())

    if tls_key:
        mode_env["ANSIBUG_TLS_SERVER_KEYFILE"] = str(tls_key.absolute())

    if tls_password:
        # FUTURE: Find a better way to share this
        mode_env["ANSIBUG_TLS_SERVER_KEY_PASSWORD"] = tls_password

    if tls_client_ca:
        mode_env["ANSIBUG_TLS_CLIENT_CA"] = str(tls_client_ca.absolute())

    _exec_playbook(
        playbook_args=playbook_args,
        mode="listen",
        mode_env=mode_env,
        addr=addr,
        no_wait=no_wait,
        use_tls=use_tls,
        log_file=log_file,
        log_level=log_level,
    )


def _exec_playbook(
    playbook_args: list[str],
    mode: t.Literal["connect", "listen"],
    mode_env: dict[str, str],
    addr: str,
    *,
    no_wait: bool = True,
    use_tls: bool = False,
    log_file: pathlib.Path | None = None,
    log_level: LogLevel = "info",
) -> None:
    """Common launch code for connect and listen modes."""
    ansible_env = os.environ | mode_env | _configure_ansible_env()
    ansible_env["ANSIBUG_DEBUG_ADDR"] = addr
    ansible_env["ANSIBUG_MODE"] = mode
    ansible_env["ANSIBUG_NO_WAIT_FOR_CONFIG_DONE"] = str(no_wait).lower()
    ansible_env["ANSIBUG_USE_TLS"] = str(use_tls)

    if log_file:
        ansible_env["ANSIBUG_LOG_FILE"] = str(log_file.absolute())
        ansible_env["ANSIBUG_LOG_LEVEL"] = log_level

    argv = [sys.executable, "-m", "ansible", "playbook"] + playbook_args

    # Needed for coverage collection in testing to run the collectors before
    # execve is called.
    # https://github.com/nedbat/coveragepy/issues/43
    if "COVERAGE_RUN" in os.environ and hasattr(atexit, "_run_exitfuncs"):
        atexit._run_exitfuncs()

    os.execve(argv[0], argv, env=ansible_env)  # pragma: nocover


def _configure_ansible_env() -> dict[str, str]:
    """Builds the new ansibug enabled config options."""
    ansibug_path = pathlib.Path(__file__).parent.absolute()

    # Get the existing config options so we can safely prepend the ansibug
    # settings without clobbering the existing ones. This should reflect the
    # config the subsequent ansible-playbook call will use as it's run in the
    # same cwd with the same env vars.
    raw_config_dump = subprocess.run(
        [
            sys.executable,
            "-m",
            "ansible",
            "config",
            "dump",
            "--format",
            "json",
        ],
        # Explicitly set verbosity so it doesn't display version info breaking
        # the json parsing.
        env=os.environ | {"ANSIBLE_VERBOSITY": "0"},
        capture_output=True,
        check=True,  # FUTURE: Maybe a nicer exception would be useful
    )
    config_dump = json.loads(raw_config_dump.stdout)

    enabled_callbacks = ["ansibug.dap.debug"]
    collections_path = [str(ansibug_path)]
    # FUTURE: Pass along the strategy used for ansibug to override.
    # strategy = "linear"

    to_check = {"CALLBACKS_ENABLED", "COLLECTIONS_PATHS"}
    for config_entry in config_dump:
        name = config_entry.get("name", None)
        if not name:
            continue

        if name == "CALLBACKS_ENABLED":
            to_check.remove("CALLBACKS_ENABLED")
            enabled_callbacks.extend(config_entry["value"])

        elif name == "COLLECTIONS_PATHS":
            to_check.remove("COLLECTIONS_PATHS")
            collections_path.extend(config_entry["value"])

        # elif name == "DEFAULT_STRATEGY":
        #     to_check.remove("DEFAULT_STRATEGY")
        #     strategy = config_entry["value"]

        if not to_check:
            break

    # Insert this module into the first PYTHONPATH so Ansible can import it.
    python_path = [e for e in os.environ.get("PYTHONPATH", "").split(os.pathsep) if e]
    python_path.insert(0, str(ansibug_path.parent))

    return {
        "ANSIBLE_CALLBACKS_ENABLED": ",".join(enabled_callbacks),
        "ANSIBLE_COLLECTIONS_PATH": os.pathsep.join(collections_path),
        "ANSIBLE_STRATEGY": "ansibug.dap.debug",
        "PYTHONPATH": os.pathsep.join(python_path),
    }



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import logging
import pathlib
import pickle
import socket
import ssl
import threading
import types
import typing as t

from ._socket_helper import (
    CancelledError,
    SocketCancellationToken,
    SocketHelper,
    create_uds_with_mask,
    parse_addr_string,
)
from .dap import ProtocolMessage

log = logging.getLogger(__name__)


class MPProtocol:
    def on_msg_received(self, msg: ProtocolMessage) -> None:
        """Called when a message has been received from the peer."""
        return  # pragma: nocover

    def connection_closed(self, exp: Exception | None) -> None:
        """Called when the connection is closed, exp will contain an exception if an error occurred."""
        return  # pragma: nocover


class MPQueue:
    def __init__(
        self,
        sock: SocketHelper,
        proto: t.Callable[[], MPProtocol],
        cancel_token: SocketCancellationToken | None = None,
    ) -> None:
        self._socket = sock
        self._proto = proto()
        self._cancel_token = cancel_token or SocketCancellationToken()
        self._recv_thread: threading.Thread | None = None

    def __enter__(self) -> MPQueue:
        return self

    def __exit__(
        self,
        exception_type: type[BaseException] | None = None,
        exception_value: BaseException | None = None,
        traceback: types.TracebackType | None = None,
        **kwargs: t.Any,
    ) -> None:
        self.stop()

    @property
    def address(self) -> str:
        if self._socket.family == socket.AddressFamily.AF_UNIX:
            host = self._socket.getsockname()
            return f"uds://{host}"

        elif self._socket.family == socket.AddressFamily.AF_INET6:
            host, port, _, _ = self._socket.getsockname()
            return f"tcp://[{host}]:{port}"

        else:
            host, port = self._socket.getsockname()
            return f"tcp://{host}:{port}"

    def send(
        self,
        data: ProtocolMessage,
    ) -> None:
        b_data = pickle.dumps(data)
        b_len = len(b_data).to_bytes(4, byteorder="little")
        self._socket.send(b_len + b_data, self._cancel_token)

    def start(
        self,
        timeout: float = 0,
    ) -> None:
        self._recv_thread = threading.Thread(
            target=self._recv_handler,
            name=f"{type(self).__name__} Recv",
        )
        self._recv_thread.start()

    def stop(self) -> None:
        self._cancel_token.cancel()
        if self._recv_thread:
            self._recv_thread.join()
        self._recv_thread = None

        self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()

    def _recv_handler(self) -> None:
        try:
            while True:
                b_data_len = self._socket.recv(4, self._cancel_token)
                if not b_data_len:
                    break

                data_len = int.from_bytes(b_data_len, byteorder="little", signed=False)
                b_data = self._socket.recv(data_len, self._cancel_token)

                py_obj = pickle.loads(b_data)
                obj = t.cast(ProtocolMessage, py_obj)
                self._proto.on_msg_received(obj)

        except CancelledError:
            self._proto.connection_closed(None)

        except Exception as e:
            self._proto.connection_closed(e)

        else:
            self._proto.connection_closed(None)


class ClientMPQueue(MPQueue):
    def __init__(
        self,
        address: str,
        proto: t.Callable[[], MPProtocol],
        *,
        ssl_context: ssl.SSLContext | None = None,
        cancel_token: SocketCancellationToken | None = None,
    ) -> None:
        # Use a blank socket, the real one is created in start()
        super().__init__(SocketHelper("client", socket.socket()), proto, cancel_token=cancel_token)
        self._address = parse_addr_string(address)
        self._ssl_context = ssl_context

    def start(
        self,
        timeout: float = 0,
    ) -> None:
        addr_info: list[
            tuple[
                socket.AddressFamily,
                socket.SocketKind,
                int,
                str,
                t.Any,
            ]
        ]
        if isinstance(self._address, str):
            hostname = self._address
            addr_info = [(socket.AF_UNIX, socket.SOCK_STREAM, -1, "", self._address)]
            # Python's ssl handling doesn't like hostnames greater than 63
            # characters. Using SSL with a UDS should only really happen in
            # testing so we can safely truncate this.
            ssl_hostname = hostname[:63]

        else:
            hostname, port = self._address
            ssl_hostname = hostname
            addr_info = socket.getaddrinfo(hostname, port, 0, socket.SOCK_STREAM)

        sock = None
        error = OSError(f"Found no results for getaddrinfo for {self._address}")
        for family, socktype, proto, _, addr in addr_info:
            sock = socket.socket(family, socktype, proto)
            try:
                self._cancel_token.connect(sock, addr, timeout=timeout)
                break

            except OSError as e:
                error = e
                sock.close()
            sock = None

        if sock:
            self._socket = SocketHelper("client", sock)
        else:
            raise error

        if self._ssl_context:
            self._socket.wrap_tls(
                self._ssl_context,
                server_side=False,
                server_hostname=ssl_hostname,
            )
            # TLS 1.3 won't validate their cert was accepted until its first
            # read. This ensures the connection is validated before it is
            # going to be used.
            self._socket.recv(1, self._cancel_token)
        else:
            # Ensures that the server is actually connected to us and we aren't
            # in the backlog waiting for an accept().
            self._socket.recv(1, self._cancel_token)

        super().start()


class ServerMPQueue(MPQueue):
    def __init__(
        self,
        address: str,
        proto: t.Callable[[], MPProtocol],
        *,
        ssl_context: ssl.SSLContext | None = None,
        cancel_token: SocketCancellationToken | None = None,
    ) -> None:
        parsed_addr = parse_addr_string(address)

        self._uds_path: str | None = None
        if isinstance(parsed_addr, str):
            self._uds_path = parsed_addr
            sock = create_uds_with_mask(parsed_addr, 0o700)
            sock.listen()

        else:
            sock_kwargs: dict[str, t.Any] = {}
            if socket.has_dualstack_ipv6():  # pragma: nocover
                # Allows the server to bind on both IPv4 and IPv6.
                sock_kwargs |= {
                    "family": socket.AF_INET6,
                    "dualstack_ipv6": True,
                }

            sock = socket.create_server(
                parsed_addr,
                # If port 0 is requested, don't set SO_REUSEPORT
                reuse_port=parsed_addr[1] != 0,
                backlog=0,
                **sock_kwargs,
            )

        super().__init__(SocketHelper("server", sock), proto, cancel_token=cancel_token)
        self._ssl_context = ssl_context

    def start(
        self,
        timeout: float = 0,
        cancel_queue: MPQueue | None = None,
    ) -> None:
        cancel_socket = None
        if cancel_queue:
            cancel_socket = cancel_queue._socket

        conn = self._socket.accept(
            self._cancel_token,
            timeout=timeout,
            cancel_socket=cancel_socket,
        )

        if self._ssl_context:
            # If the server fails to complete the TLS handshake we don't want
            # to shutdown the socket, instead this will recreate it allowing a
            # new client to connect. This allows an attach request to be
            # retried if the server mandates a certificate but the client does
            # not offer one (or is rejected).
            while True:
                try:
                    conn.wrap_tls(
                        self._ssl_context,
                        server_side=True,
                    )
                    # The client will read a single byte after the handshake to
                    # verify it was able to auth correctly. TLS 1.3 only errors
                    # on the client side on the first recv/send call after the
                    # handshake.
                    conn.send(b"\x00", self._cancel_token)

                except ssl.SSLError:
                    log.exception("Server TLS handshake failed, trying again")
                    conn.close()
                    conn = self._socket.accept(
                        self._cancel_token,
                        timeout=timeout,
                        cancel_socket=cancel_socket,
                    )

                else:
                    break

        else:
            # The client will read a single byte after the connection to verify
            # it is actually connected to a server and not just part of the
            # backlog.
            conn.send(b"\x00", self._cancel_token)

        # The original server socket is no longer needed, replace it with the
        # client connection one.
        self._socket.close()
        self._socket = conn

        super().start()

    def stop(self) -> None:
        if self._uds_path:
            pathlib.Path(self._uds_path).unlink(missing_ok=True)

        super().stop()



# Copyright (c) 2023 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import argparse
import dataclasses
import shlex
import typing as t
from typing import IO


class ReturnErrorMessage(Exception): ...


class ErrorCapturingParser(argparse.ArgumentParser):
    def error(self, message: str) -> t.NoReturn:
        raise ReturnErrorMessage(message)

    def print_help(self, file: IO[str] | None = None) -> None:
        raise ReturnErrorMessage(self.format_help())


@dataclasses.dataclass(frozen=True)
class ReplCommand:
    command: t.Literal["remove_option", "set_option", "set_hostvar", "template"]


@dataclasses.dataclass(frozen=True)
class TemplateCommand(ReplCommand):
    expression: str


@dataclasses.dataclass(frozen=True)
class RemoveVarCommand(ReplCommand):
    name: str


@dataclasses.dataclass(frozen=True)
class SetVarCommand(ReplCommand):
    name: str
    expression: str


def parse_repl_args(
    args: str,
) -> ReplCommand | str:
    """Parses an ansibug repl console command.

    Parses the raw command from the ansibug repl evaluation request. It uses
    argparse to parse the string value into a known command and their arguments
    otherwise it returns a string to send back to the client.

    Args:
        args: The raw string to parse.

    Returns:
        ReplCommand | str: The ReplCommand instance if it's a valid command,
        otherwise the string to display back to the client.
    """
    aliases = {
        "remove_option": ["ro"],
        "set_option": ["so"],
        "set_hostvar": ["sh"],
        "template": ["t"],
    }

    parser = ErrorCapturingParser(
        prog="!",
        description="Ansibug debug console repl commands",
    )

    command_parser = parser.add_subparsers(
        dest="command",
        required=True,
        help="The command to run",
    )

    remove_option = command_parser.add_parser(
        "remove_option",
        aliases=["ro"],
        description="Removes the option specified from the current module options.",
        help="Remove a module option.",
    )
    remove_option.add_argument(
        "name",
        help="The name of the option to remove.",
    )

    set_option = command_parser.add_parser(
        "set_option",
        aliases=["so"],
        description="Adds or sets the provided option for the current module options using the expression provided.",
        help="Set a module option.",
    )
    set_option.add_argument(
        "name",
        help="The name of the option to add/set.",
    )

    set_hostvar = command_parser.add_parser(
        "set_hostvar",
        aliases=["sh"],
        description="Adds or sets the provided variable for the current host using the expression provided.",
        help="Set a host variable.",
    )
    set_hostvar.add_argument(
        "name",
        help="The name of the hostvar to add/set.",
    )

    command_parser.add_parser(
        "template",
        aliases=["t"],
        description="Templates the expression provided after the command and returns the result. The expression "
        "provided should not be wrapped in {{ ... }} as the value is already treated as an expression.",
        help="Template an expression",
    )

    # The provided value from the repl is just a string, we need to convert it
    # to a command line and parse the known args. We only do the known args
    # because some commands need the raw value provided outside of a cmd split
    # processing argparse uses.
    split_args = shlex.split(args)

    try:
        parsed_args = parser.parse_known_args(split_args)[0]
    except ReturnErrorMessage as e:
        return str(e)

    command = str(parsed_args.command)
    if command not in aliases:
        # Need to resolve the alias back to the expected command name.
        for final, cmd_aliases in aliases.items():  # pragma: nocover
            for a in cmd_aliases:
                if command == a:
                    parsed_args.command = final
                    break

            if command != parsed_args.command:
                break

    if parsed_args.command == "template":
        # The value after the command needs to be extracted.
        expr = args[args.index(command) + len(command) :].lstrip()

        return TemplateCommand(command="template", expression=expr)

    elif parsed_args.command in ["set_option", "set_hostvar"]:
        # The value after the option/var name needs to be extracted.
        remaining = args[args.index(command) + len(command) :].lstrip()
        remaining = remaining[remaining.index(parsed_args.name) + len(parsed_args.name) :]
        if remaining.startswith('"') or remaining.startswith("'"):
            remaining = remaining[1:]

        expr = remaining.lstrip()

        return SetVarCommand(parsed_args.command, parsed_args.name, expr)

    else:
        return RemoveVarCommand(parsed_args.command, parsed_args.name)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import argparse
import os
import pathlib
import sys

from ._debug_adapter import start_dap
from ._exec_playbook import exec_playbook_connect, exec_playbook_listen

HAS_ARGCOMPLETE = True
try:
    import argcomplete
except ImportError:  # pragma: nocover
    HAS_ARGCOMPLETE = False


def _add_launch_common_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--no-wait",
        action="store_true",
        help="Do not wait for the client to connect and send the "
        "configurationDone request before starting the playbook.",
    )

    parser.add_argument(
        "playbook_args",
        nargs=argparse.REMAINDER,
        help="Arguments to use when launching ansible-playbook.",
    )


def _add_log_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--log-file",
        action="store",
        type=_parse_path,
        help="Enable file logging to the file at this path.",
    )

    parser.add_argument(
        "--log-level",
        action="store",
        choices=["info", "debug", "warning", "error"],
        default="info",
        type=str,
        help="Set the logging filter level of the logger when --log-file is set. Defaults to info",
    )


def _add_tls_server_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--tls-cert",
        action="store",
        type=_parse_path,
        help="The path to the TLS server certificate pem. This can either be "
        "the certificate by itself or a bundle of the certificate and key as "
        "a PEM file. Use --tls-key to specify a separate key file and "
        "--tls-key-pass if the key is password protected. If no certificate "
        "is specified, the server will exchange data without TLS.",
    )

    parser.add_argument(
        "--tls-key",
        action="store",
        type=_parse_path,
        help="The path to the TLS server certificate key pem. This can be "
        "used if the --tls-cert only contains the certificate and the key is "
        "located in another file. Use --tls-key-pass to specify the password "
        "to decrypt the key if needed.",
    )

    parser.add_argument(
        "--tls-key-pass",
        action="store",
        default=os.environ.get("ANSIBUG_TLS_KEY_PASS", None),
        type=str,
        help="The password for the TLS key if it is encrypted. The "
        "environment variable ANSIBUG_TLS_KEY_PASS can also be used to "
        "provide the password without passing it through the command line.",
    )

    parser.add_argument(
        "--tls-client-ca",
        action="store",
        type=_parse_path,
        help="The path to a TLS CA bundle file or directory to use with "
        "verifying the identity of the client. If set the client must "
        "provide a certificate signed by a CA in the bundle specified for it "
        "to connect. There are no checks on the client provided cert's key "
        "usage or extended key usage, just the isssuer. If not set then no "
        "client authentication is needed.",
    )


def _parse_path(path: str) -> pathlib.Path:
    return pathlib.Path(os.path.expanduser(os.path.expandvars(path)))


def parse_args(
    args: list[str],
) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="python -m ansibug",
        description="Ansible Debug Adapter Protocol Launcher (DAP).",
    )

    action = parser.add_subparsers(
        dest="action",
        required=True,
        help="The action for ansibug to perform",
    )

    # Connect

    connect = action.add_parser(
        "connect",
        description="Has the new ansible-playbook process start in connect "
        "mode. The connect mode will have the new process connect to an "
        "existing DA server that has set up a listener. The addr must be "
        "specified for the ansible-playbook process to connect to.",
        help="Launch a debuggable ansible-playbook process in connect mode.",
    )

    connect.add_argument(
        "--addr",
        action="store",
        type=str,
        required=True,
        help="Have the ansible-playbook process connect to this socket addr that is bound to the DAP server.",
    )
    _add_log_args(connect)
    _add_launch_common_args(connect)

    # DAP

    dap = action.add_parser(
        "dap",
        description="Start a new Debug Adapter Protocol process for a client "
        "to communicate with over stdin and stdout. This process is in "
        "charge of interacting with the DAP client and coordinating the debug "
        "session. All relevant debug information is passed in through stdin "
        "using the DAP protocol messages.",
        help="Start an executable Debug Adapter Protocol process",
    )
    _add_log_args(dap)

    # Listen

    listen = action.add_parser(
        "listen",
        description="Has the new ansible-playbook process start in listen "
        "mode. The listen mode allows the debug client to attach to the "
        "playbook later on through an attach request either locally by the "
        "playbook process id or through the host:port it is listening on.",
        help="Launch a debuggable ansible-playbook process in listen mode.",
    )

    listen.add_argument(
        "--addr",
        action="store",
        default="uds://",
        type=str,
        help="Specify the socket address to listen on. The address can "
        "either be a TCP or UDS address in the format 'tcp://hostname:port' "
        "or 'uds:///path/to/uds'. The default behavior is to create a UDS "
        "socket that only allows the current user to connect. Using a TCP "
        "address can allow connections from another host on that socket. For "
        "example 'tcp://:0' will bind a random port to all IPs on the "
        "current host. Using 'tcp://127.0.0.1:0' will bind a random port for "
        "localhost communication. Using the address 'uds://' will create a "
        "random Unix Domain Socket automatically. The address used will be "
        "displayed on the console when it is available.",
    )
    _add_log_args(listen)
    _add_launch_common_args(listen)
    listen.add_argument(
        "--wrap-tls",
        action="store_true",
        help="Has the ansible-playbook process wrap its socket communication with the DA server with TLS",
    )
    _add_tls_server_args(listen)

    if HAS_ARGCOMPLETE:  # pragma: nocover
        argcomplete.autocomplete(parser)

    return parser.parse_args(args)


def main() -> None:
    args = parse_args(sys.argv[1:])

    if args.action == "dap":
        start_dap(args.log_file, args.log_level)
        return

    if args.action == "connect":
        exec_playbook_connect(
            args.playbook_args,
            addr=args.addr,
            no_wait=args.no_wait,
            log_file=args.log_file,
            log_level=args.log_level,
        )

    else:
        exec_playbook_listen(
            args.playbook_args,
            addr=args.addr,
            no_wait=args.no_wait,
            use_tls=args.wrap_tls,
            tls_cert=args.tls_cert,
            tls_key=args.tls_key,
            tls_password=args.tls_key_pass,
            tls_client_ca=args.tls_client_ca,
            log_file=args.log_file,
            log_level=args.log_level,
        )


if __name__ == "__main__":  # pragma: nocover
    main()



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import base64
import collections.abc
import contextlib
import errno
import logging
import os
import os.path
import pathlib
import select
import socket
import ssl
import tempfile
import threading
import typing as t
import urllib.parse
import uuid

log = logging.getLogger(__name__)


def parse_addr_string(
    value: str,
) -> t.Any:
    """Parse the an address string.

    Parses the address string using the formats known to Ansibug. Currently the
    following formats are allowed:

        # TCP for all interfaces and available port
        tcp://

        # Hostname
        tcp://hostname:1234

        # IPv4
        tcp://192.168.1.0:5678

        # IPv6
        tcp://[::1]:1010

        # No hostname (represents 0.0.0.0)
        tcp://:2020

        # UDS generate random filename in tmp
        uds://

        # UDS with just the filename
        uds://ansibug-dap-sock

        # UDS with full path
        uds:/tmp/ansibug-dap-socket

        # UDS with full path alternative
        uds:///tmp/ansibug-dap-socket

    A TCP socket must include both the hostname and port. A UDS socket can be
    just the socket filename located in the temporary directory or the full
    path to the UDS socket.

    Args:
        value: The string to parse.

    Returns:
        Any: The required address for the scheme specified.
    """
    value_split = urllib.parse.urlsplit(value)

    if value_split.scheme == "tcp":
        if not value_split.netloc and not value_split.port:
            return "", 0

        if value_split.port is not None:
            return value_split.hostname or "", value_split.port

        else:
            raise ValueError("Specifying a tcp address must include the port")

    elif value_split.scheme == "uds":
        if value_split.netloc and value_split.path:
            raise ValueError("Specifying a uds address must only contain the port or netloc, not both")

        elif value_split.netloc:
            return os.path.join(tempfile.gettempdir(), value_split.netloc)

        elif value_split.path:
            return value_split.path

        else:
            return str(pathlib.Path(tempfile.gettempdir()) / f"ansibug-dap-{uuid.uuid4()}")

    else:
        raise ValueError(f"An address must be for the tcp or uds scheme")


def create_uds_with_mask(
    addr: str,
    mask: int,
) -> socket.socket:
    """Create a server Unix Domain Socket with mask specified.

    This function can be used to create a bound Unix Domain Socket with the
    mask specified.

    Args:
        addr: The UDS address to bind to.
        mask: The mask to set on the UDS created.

    Returns:
        socket.socket: The bound UDS that was created.
    """
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    original_umask = None
    try:
        try:
            # Works on Linux only
            os.fchmod(sock.fileno(), mask)
        except OSError as e:
            if e.errno != errno.EINVAL:  # pragma: nocover
                raise

            # fchmod doesn't work on other platforms like macOS. At this point
            # no thread should have started so it should be safe to use the
            # umask before binding the socket.
            original_umask = os.umask(~mask)

        sock.bind(addr)

    finally:
        if original_umask is not None:
            os.umask(original_umask)

    return sock


class SocketHelper:
    def __init__(
        self,
        use: str,
        sock: socket.socket,
    ) -> None:
        self.use = use
        self._sock = sock

    @property
    def family(self) -> socket.AddressFamily:
        return self._sock.family

    def close(self) -> None:
        log.debug("Closing %s socket", self.use)
        self._sock.close()

    def accept(
        self,
        cancel_token: SocketCancellationToken,
        timeout: float = 0,
        cancel_socket: SocketHelper | None = None,
    ) -> SocketHelper:
        log.debug("Socket %s starting accept", self.use)
        conn, addr = cancel_token.accept(
            self._sock,
            timeout=timeout,
            cancel_socket=cancel_socket._sock if cancel_socket else None,
        )
        log.debug("Socket %s accepted conn from '%s'", self.use, addr)

        return SocketHelper(self.use, conn)

    def getsockname(self) -> t.Any:
        return self._sock.getsockname()

    def recv(
        self,
        n: int,
        cancel_token: SocketCancellationToken,
    ) -> bytes:
        """Wraps recv but ensures the data length specified is read."""
        buffer = bytearray(n)
        view = memoryview(buffer)
        read = 0

        while read < n:
            data_read = cancel_token.recv_into(self._sock, view[read:], n - read)
            read += data_read

            # On a socket shutdown 0 bytes will be read.
            if data_read == 0:
                break

        data = bytes(buffer[:read])
        if log.isEnabledFor(logging.DEBUG):
            log.debug("Socket %s recv(%d): '%s'", self.use, n, base64.b64encode(data).decode())
        return data

    def send(
        self,
        data: bytes,
        cancel_token: SocketCancellationToken,
    ) -> None:
        """Wraps send but ensures all the data is sent."""
        if log.isEnabledFor(logging.DEBUG):
            log.debug("Socket %s send: '%s'", self.use, base64.b64encode(data).decode())
        cancel_token.sendall(self._sock, data)

    def shutdown(
        self,
        how: int,
    ) -> None:
        try:
            self._sock.shutdown(how)
        except OSError:
            pass

    def wrap_tls(
        self,
        ssl_context: ssl.SSLContext,
        *,
        server_side: bool = True,
        server_hostname: str | None = None,
    ) -> None:
        log.debug("Wrapping socket with TLS context")
        self._sock = ssl_context.wrap_socket(
            self._sock,
            server_side=server_side,
            server_hostname=server_hostname,
        )


class SocketCancellationToken:
    def __init__(self) -> None:
        self._cancel_funcs: dict[int, t.Callable[[], None]] = {}
        self._cancel_id = 0
        self._cancelled = False
        self._lock = threading.Lock()

    def accept(
        self,
        sock: socket.socket,
        timeout: float = 0,
        cancel_socket: socket.socket | None = None,
    ) -> tuple[socket.socket, t.Any]:
        def _cancel() -> None:
            # Using shutdown(socket.SHUT_RDWR) fails on macOS with 'Socket is
            # not connected'. Fallback to close() if the shutdown failed.
            try:
                sock.shutdown(socket.SHUT_RDWR)
            except OSError:
                sock.close()

        with self.with_cancel(_cancel):
            try:
                # When cancelled, select will detect that sock is ready for a
                # read and accept() will raise OSError. In the rare event the
                # socket was connected to and closed/shutdown between select
                # and the subsequent recv/send on the socket will act like it's
                # disconnected
                socks = [sock]
                if cancel_socket:
                    socks.append(cancel_socket)

                rd, _, _ = select.select(socks, [], [], timeout or None)
                if not rd:
                    raise TimeoutError("Timed out waiting for socket.accept()")
                elif cancel_socket and cancel_socket in rd:
                    raise CancelledError()

                return rd[0].accept()

            except OSError:
                # On macOS, the socket fd is closed on cancel
                if self._cancelled:
                    raise CancelledError()
                else:
                    raise

    def connect(
        self,
        sock: socket.socket,
        addr: t.Any,
        timeout: float = 0,
    ) -> None:
        with self.with_cancel(lambda: sock.shutdown(socket.SHUT_RDWR)):
            # There seems to be a bug in some Linux kernel versions where
            # trying to shutdown a socket during a connect() fails with EBUSY.
            # This was reproduced in Ubuntu 22.04 (Linux 5.15.0) but could not
            # be reproduced in Fedora 38 (Linux 6.2) or Arch at Linux 6.5.9.
            # Using select is an option to wait until the socket is writable.
            sock.setblocking(False)

            try:
                sock.connect(addr)
            except BlockingIOError:
                pass

            _, wd, _ = select.select([], [sock], [], timeout or None)
            if not wd:
                raise TimeoutError("Timed out waiting for socket.connect")
            elif self._cancelled:
                raise CancelledError()
            elif err := sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR):
                raise OSError(err, os.strerror(err))

            sock.setblocking(True)

    def recv_into(
        self,
        sock: socket.socket,
        buffer: bytearray | memoryview,
        n: int,
    ) -> int:
        with self.with_cancel(lambda: sock.shutdown(socket.SHUT_RD)):
            res = sock.recv_into(buffer, n)
            if self._cancelled:
                raise CancelledError()

            return res

    def sendall(
        self,
        sock: socket.socket,
        data: bytes,
    ) -> None:
        with self.with_cancel(lambda: sock.shutdown(socket.SHUT_WR)):
            try:
                sock.sendall(data)
            except OSError:
                if self._cancelled:
                    raise CancelledError()
                else:
                    raise

    def cancel(self) -> None:
        with self._lock:
            self._cancelled = True

            for cancel_id, func in self._cancel_funcs.items():
                log.debug("Cancelling function with id %d", cancel_id)
                func()

            self._cancel_funcs = {}

    @contextlib.contextmanager
    def with_cancel(
        self,
        cancel_func: t.Callable[[], None],
    ) -> collections.abc.Generator[None, None, None]:
        with self._lock:
            if self._cancelled:
                raise CancelledError()

            cancel_id = self._cancel_id
            self._cancel_id += 1
            self._cancel_funcs[cancel_id] = cancel_func

        try:
            log.debug("Calling cancellable function with id %d", cancel_id)
            yield

        finally:
            with self._lock:
                self._cancel_funcs.pop(cancel_id, None)


class CancelledError(Exception):
    pass



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import base64
import dataclasses
import functools
import json
import logging
import os
import pathlib
import shlex
import sys
import tempfile
import traceback
import types
import typing as t

from . import dap as dap
from ._debuggee import (
    DebugConfiguration,
    PathMapping,
    PlaybookProcessInfo,
    get_pid_info_path,
)
from ._logging import LogLevel, configure_file_logging
from ._mp_queue import ClientMPQueue, MPProtocol, MPQueue, ServerMPQueue
from ._socket_helper import CancelledError
from ._tls import create_client_tls_context

log = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class AttachArguments:
    """Arguments for AttachRequest.

    These are the known arguments for attach request. See the from_json method
    to see the default values. There are two ways to specify a process for
    attach; through the local process id or with the manual socket address and
    port.

    If the process_id is specified, the code will attempt to lookup the
    connection details of that process stored in a local temporary file under
    the PID identifier.

    If the address and port is specified, ensure use_tls is set to whether the
    socket server is expecting TLS or not.

    If TLS client authentication is required by the server the tls_cert must
    be set to the path of a PEM encoded certificate. This file can include both
    the certificate and key, if the key is separate use tls_key to specify that
    file.

    Args:
        process_id: The local ansible-playbook process to attach to.
        address: The socket address to connect to.
        connect_timeout: The timeout in seconds for the debug adapter to wait
            for when attempting to connect to the ansible-playbook process.
        use_tls: Whether to use TLS or not.
        tls_verification: The TLS verification settings.
        tls_cert: The certificate to use for client authentication if required
            by the server.
        tls_key: The certificate key if stored in a separate file.
        tls_key_password: The password for the certificate key if needed.
        path_mappings: A list of paths to map between a local and remote root.
    """

    process_id: t.Optional[int]
    address: str
    connect_timeout: float
    use_tls: bool
    tls_verification: t.Union[t.Literal["verify", "ignore"], str]
    tls_cert: t.Optional[str]
    tls_key: t.Optional[str]
    tls_key_password: t.Optional[str]
    path_mappings: t.List[PathMapping]

    def get_connection_tuple(self) -> tuple[str, bool]:
        """Gets the address and use_tls settings for this request."""
        if self.process_id is not None:
            pid_path = get_pid_info_path(self.process_id)
            if not pid_path.exists():
                raise ValueError(f"Failed to find process pid file at '{pid_path}'")

            proc_json = json.loads(pid_path.read_text())
            proc_info = PlaybookProcessInfo.from_json(proc_json)

            return (proc_info.address, proc_info.use_tls)

        elif self.address:
            return (self.address, self.use_tls)

        else:
            raise ValueError("Expected processId or address/port to be specified for attach.")

    @classmethod
    def from_json(
        cls,
        data: dict[str, t.Any],
    ) -> AttachArguments:
        path_mappings = []
        for mapping in data.get("pathMappings", []):
            path_mappings.append(
                PathMapping(
                    local_root=mapping["localRoot"],
                    remote_root=mapping["remoteRoot"],
                )
            )

        return AttachArguments(
            process_id=data.get("processId", None),
            address=data.get("address", ""),
            connect_timeout=float(data.get("connectTimeout", 5.0)),
            use_tls=bool(data.get("useTls", False)),
            tls_verification=data.get("tlsVerification", "verify"),
            tls_cert=data.get("tlsCertificate", None),
            tls_key=data.get("tlsKey", None),
            tls_key_password=data.get("tlsKeyPassword", None),
            path_mappings=path_mappings,
        )


@dataclasses.dataclass(frozen=True)
class LaunchArguments:
    """Arguments for LaunchRequest.

    These are the known arguments for launch request. See the from_json method
    to see the default values. A launch request will spawn a new
    ansible-playbook process through 'python -m ansibug connect ...' which sets
    up the new process to talk to the debug adapter server.

    Args:
        playbook: The playbook to launch.
        args: Extra arguments to run with ansible-playbook.
        cwd: The working directory to launch ansible-playbook with.
        env: Extra environment variables to use when launching the process.
        console: The console type to use.
        connect_timeout: The timeout in seconds to wait for the newly spawned
            ansible-playbook process to connect back to the debug adapter.
        path_mappings: A list of paths to map between a local and remote root.
        log_file: Path to a local file to log ansibug details to.
        log_level: The level of logging to use.
    """

    playbook: str
    args: t.List[str]
    cwd: str
    env: t.Dict[str, t.Optional[str]]
    console: t.Literal["integrated", "external"]
    connect_timeout: float
    path_mappings: t.List[PathMapping]
    log_file: t.Optional[str]
    log_level: LogLevel

    @classmethod
    def from_json(
        cls,
        data: dict[str, t.Any],
    ) -> LaunchArguments:
        if not (playbook := data.get("playbook", None)):
            raise ValueError("Expected playbook to be specified for launch.")

        console = data.get("console", "integratedTerminal")
        console_kind: t.Literal["external", "integrated"]
        if console == "integratedTerminal":
            console_kind = "integrated"
        elif console == "externalTerminal":
            console_kind = "external"
        else:
            raise ValueError(f"Unknown console value '{console}' - expected integratedTerminal or externalTerminal.")

        path_mappings = []
        for mapping in data.get("pathMappings", []):
            path_mappings.append(
                PathMapping(
                    local_root=mapping["localRoot"],
                    remote_root=mapping["remoteRoot"],
                )
            )

        return LaunchArguments(
            playbook=playbook,
            args=data.get("args", []),
            cwd=data.get("cwd", ""),
            env=data.get("env", {}),
            console=console_kind,
            connect_timeout=float(data.get("connectTimeout", 5.0)),
            path_mappings=path_mappings,
            log_file=data.get("logFile", None),
            log_level=data.get("logLevel", "info"),
        )


@dataclasses.dataclass
class _RunInTerminalDetails:
    request: dap.LaunchRequest
    debuggee: ServerMPQueue
    arguments: LaunchArguments
    launch_queue: ServerMPQueue


def start_dap(
    log_file: pathlib.Path | None = None,
    log_level: LogLevel = "info",
) -> None:
    if log_file:
        configure_file_logging(str(log_file.absolute()), log_level)

    log.info("DAP starting")
    try:
        with DAServer() as da:
            da.start()
    except:
        log.exception("Exception when running DA Server")
        raise

    log.info("DAP ending")


class DAProtocol(MPProtocol):
    def __init__(
        self,
        server: DAServer,
    ) -> None:
        self._server = server

    def on_msg_received(
        self,
        msg: dap.ProtocolMessage,
    ) -> None:
        self._server.send_to_client(msg)

    def connection_closed(
        self,
        exp: Exception | None,
    ) -> None:
        self._server.stop(exp, from_debuggee=True)


class DAServer:
    def __init__(self) -> None:
        self._adapter = dap.DebugAdapterConnection()

        self._proto = DAProtocol(self)
        self._debuggee: MPQueue | None = None
        self._launch_queue: MPQueue | None = None
        self._incoming_requests: dict[int, dap.Request] = {}
        self._run_in_response_data: dict[int, _RunInTerminalDetails] = {}
        self._terminated_sent = False

    def __enter__(self) -> DAServer:
        return self

    def __exit__(
        self,
        exception_type: type[BaseException] | None = None,
        exception_value: BaseException | None = None,
        traceback: types.TracebackType | None = None,
        **kwargs: t.Any,
    ) -> None:
        self.stop(exp=exception_value)

    def start(self) -> None:
        """Start DA Server.

        Starts the DA server and continues to read from stdin for messages sent
        by the client. This continues until the client has sent the disconnect
        message.
        """
        stdin = sys.stdin.buffer.raw  # type: ignore[union-attr]
        adapter = self._adapter

        is_connected = True
        while is_connected:
            data = stdin.read(4096)

            if log.isEnabledFor(logging.DEBUG):
                log.debug("STDIN: %s", data.decode())
            adapter.receive_data(data)

            while msg := adapter.next_message():
                # Trace each request from the client so we can return an
                # ErrorMessage if there was a critical failure.
                if isinstance(msg, dap.Request):
                    self._incoming_requests[msg.seq] = msg

                if isinstance(msg, dap.Request) and self._debuggee:
                    if isinstance(msg, dap.DisconnectRequest):
                        is_connected = False
                        self._send_disconnect(msg, self._debuggee)
                    else:
                        self._debuggee.send(msg)

                else:
                    self._process_msg(msg)

    def stop(
        self,
        exp: BaseException | None = None,
        from_debuggee: bool = False,
    ) -> None:
        """Stops the debuggee connection.

        This stops the debuggee connection and marks all the outstanding
        messages as done.

        Args:
            exp: If this is being stopped for an error, this is the exception
                details.
            from_debuggee: The stop is being called as the debuggee has
                disconnected.
        """
        # Ensure the DisconnectResponse is sent if there is an outstanding
        # request.
        for req in list(self._incoming_requests.values()):
            if isinstance(req, dap.DisconnectRequest):
                self.send_to_client(dap.DisconnectResponse(request_seq=req.seq))

        if exp:
            # For every incoming request relay the exception back to the client
            # for it to display to the end user.
            for req in list(self._incoming_requests.values()):
                resp = dap.ErrorResponse(
                    command=req.command,
                    request_seq=req.seq,
                    message=f"Critical DAServer exception received\n{traceback.format_exc()}",
                )
                self.send_to_client(resp)

        if not from_debuggee and self._debuggee:
            self._debuggee.stop()
            self._debuggee = None

        if self._launch_queue:
            self._launch_queue.stop()
            self._launch_queue = None

        if not self._terminated_sent:
            # Both the debuggee or start() end will hit the code here. We only
            # need to send one TerminatedEvent() for whichever was first.
            self.send_to_client(dap.TerminatedEvent())
            self._terminated_sent = True

    def send_to_client(
        self,
        msg: dap.ProtocolMessage,
    ) -> int:
        stdout = sys.stdout.buffer

        req_no = self._adapter.queue_msg(msg)
        data = self._adapter.data_to_send() or b""
        if log.isEnabledFor(logging.DEBUG):
            log.debug("STDOUT: %s", data.decode())

        stdout.write(data)
        stdout.flush()

        if isinstance(msg, dap.Response):
            self._incoming_requests.pop(msg.request_seq, None)

        return req_no

    def _send_disconnect(
        self,
        msg: dap.DisconnectRequest,
        debuggee: MPQueue,
    ) -> None:
        try:
            debuggee.send(msg)
        except OSError:
            # The debuggee might have been disconnected during the send.
            pass

        # No response is sent back to the client, is it done as part of the
        # shutdown to ensure it's sent before the TerminatedEvent.

    @functools.singledispatchmethod
    def _process_msg(self, msg: dap.ProtocolMessage) -> None:
        # This should never happen.
        raise NotImplementedError(f"Debug Adapter does not support the {type(msg).__name__} message")

    @_process_msg.register
    def _(self, msg: dap.InitializeRequest) -> None:
        self.send_to_client(
            dap.InitializeResponse(
                request_seq=msg.seq,
                capabilities=dap.Capabilities(
                    supports_clipboard_context=True,
                    supports_conditional_breakpoints=True,
                    supports_configuration_done_request=True,
                    supports_set_variable=True,
                    # We could support suspend for debug in an attach but
                    # launch is harder as nothing can reattach to it. I'm not
                    # aware of any way to set this conditionally for the
                    # request so it's kept False for now.
                    supports_suspend_debuggee=False,
                    supports_terminate_debuggee=True,
                    supports_terminate_request=True,
                ),
            )
        )

    @_process_msg.register
    def _(self, msg: dap.AttachRequest) -> None:
        attach_args = AttachArguments.from_json(msg.arguments)
        addr, use_tls = attach_args.get_connection_tuple()

        ssl_context = None
        if use_tls:
            ssl_context = create_client_tls_context(
                attach_args.tls_verification,
                certfile=attach_args.tls_cert,
                keyfile=attach_args.tls_key,
                password=attach_args.tls_key_password,
            )

        self._debuggee = ClientMPQueue(
            addr,
            proto=lambda: self._proto,
            ssl_context=ssl_context,
        )

        self._debuggee.start(timeout=attach_args.connect_timeout)
        self.send_to_client(dap.AttachResponse(request_seq=msg.seq))
        self._debuggee.send(
            dap.OutputEvent(
                category=DebugConfiguration.OUTPUT_CATEGORY,
                output="",
                data=DebugConfiguration(path_mappings=attach_args.path_mappings),
            )
        )

    @_process_msg.register
    def _(self, msg: dap.LaunchRequest) -> None:
        launch_args = LaunchArguments.from_json(msg.arguments)

        self._debuggee = ServerMPQueue(
            "uds://",
            proto=lambda: self._proto,
        )

        ansibug_args = [
            "-m",
            "ansibug",
            "connect",
            "--addr",
            self._debuggee.address,
        ]
        if launch_args.log_file is not None:
            ansibug_args.extend(
                [
                    "--log-file",
                    launch_args.log_file,
                    "--log-level",
                    launch_args.log_level,
                ]
            )

        # This is used as an indicator to see whether the launched process has
        # ended or not before it had a chance to connect to our real socket. We
        # use a UDS because we can use select on it unlike files.
        self._launch_queue = ServerMPQueue("uds://", proto=MPProtocol)

        # A temp script is used that will clean itself up as well as signal
        # our launch_queue that it has exited. This has 2 benefits, the user
        # doesn't see the inner workings of ansibug and the
        # RunInTerminalResponse has a way of detecting if the shell script or
        # ansible failed or not before a timeout is hit.
        launch_script = """#!/bin/sh
set -e

cleanup () {{
    SN="{sock_name}"
    {python} -c "
import base64
import socket

n = base64.b64decode('${{SN}}').decode()
try:
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
        s.connect(n)
except OSError:
    pass
"
}}

trap cleanup EXIT INT ABRT KILL TERM

rm -f "$0"
{python} {ansibug_args} "$@"
""".format(
            sock_name=base64.b64encode(self._launch_queue.address[6:].encode()).decode(),
            python=shlex.quote(sys.executable),
            ansibug_args=" ".join(shlex.quote(c) for c in ansibug_args),
        )

        # Default permissions are 0o600, we need the executable bit.
        launch_fd, launch_file = tempfile.mkstemp(prefix="ansibug-launch-")
        os.fchmod(launch_fd, 0o700)
        with os.fdopen(launch_fd, mode="w") as launch_writer:
            launch_writer.write(launch_script)

        log.debug("Created launch script at '%s' with\n%s", launch_file, launch_script)

        playbook_args = [launch_file, launch_args.playbook]
        playbook_args.extend(launch_args.args)

        req_no = self.send_to_client(
            dap.RunInTerminalRequest(
                cwd=launch_args.cwd,
                kind=launch_args.console,
                args=playbook_args,
                title="Ansible Debug Console",
                env=launch_args.env,
            )
        )
        # The remaining work is done in the RunInTerminalResponse from the
        # client, this stores the data needed for that to work.
        self._run_in_response_data[req_no] = _RunInTerminalDetails(
            request=msg,
            debuggee=self._debuggee,
            arguments=launch_args,
            launch_queue=self._launch_queue,
        )

    @_process_msg.register
    def _(self, msg: dap.RunInTerminalResponse) -> None:
        details = self._run_in_response_data.pop(msg.request_seq)

        try:
            details.debuggee.start(
                timeout=details.arguments.connect_timeout,
                cancel_queue=details.launch_queue,
            )

        except CancelledError:
            log.debug("Launch cancel socket was connected, launched process has ended")
            self.stop()

        else:
            log.debug("Ansible process is connected to debug adapter")
            details.launch_queue.stop()
            self._launch_queue = None

            self.send_to_client(dap.LaunchResponse(request_seq=details.request.seq))
            details.debuggee.send(
                dap.OutputEvent(
                    category=DebugConfiguration.OUTPUT_CATEGORY,
                    output="",
                    data=DebugConfiguration(path_mappings=details.arguments.path_mappings),
                )
            )



# Copyright (c) 2023 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import os.path
import pathlib
import ssl
import typing as t


def create_client_tls_context(
    verify: t.Literal["verify", "ignore"] | str = "verify",
    certfile: str | None = None,
    keyfile: str | None = None,
    password: str | None = None,
) -> ssl.SSLContext:
    """Creates a client TLS context.

    Creates a TLS context with the verification settings specified. The verify
    argument can be set to verify (default), ignore to ignore any verification
    settings, or a path to a file or directory containing the CA certs to use
    for the verification process.

    Args:
        verify: The verification settings for the TLS context.
        certfile: The path to the certificate and optional key used for the
            client's identity.
        keyfile: The path to the key if not present in the certfile path.
        password: The password used to decrypt the key if it is encrypted.
        ca_trust: Optional CA file that turns on client certificate
            authentication with a cert signed by a CA in this path.

    Returns:
        ssl.SSLContext: The configured client TLS context.

    Raises:
        ValueError: The verify location path does not exist.
    """
    ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    if verify == "ignore":
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.VerifyMode.CERT_NONE

    elif verify != "verify":
        _load_verify_locations(ssl_context, verify)

    if certfile:
        ssl_context.load_cert_chain(
            certfile,
            keyfile=keyfile,
            password=password,
        )

    return ssl_context


def create_server_tls_context(
    certfile: str,
    keyfile: str | None = None,
    password: str | None = None,
    ca_trust: str | None = None,
) -> ssl.SSLContext:
    """Creates a server TLS context.

    Creates a TLS context with the certificate settings specified. If no
    certfile is specified, the platforms defaults are used.

    Args:
        certfile: The path to the certificate and optional key used for the
            server's identity.
        keyfile: The path to the key if not present in the certfile path.
        password: The password used to decrypt the key if it is encrypted.
        ca_trust: Optional CA file that turns on client certificate
            authentication with a cert signed by a CA in this path.

    Returns:
        ssl.SSLContext: The configured server TLS context.
    """
    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(
        certfile,
        keyfile=keyfile,
        password=password,
    )

    if ca_trust:
        ssl_context.verify_mode = ssl.VerifyMode.CERT_REQUIRED
        _load_verify_locations(ssl_context, ca_trust)

    return ssl_context


def _load_verify_locations(
    ssl_context: ssl.SSLContext,
    location: str,
) -> None:
    ca_path = pathlib.Path(os.path.expanduser(os.path.expandvars(location)))

    if ca_path.is_dir():
        ssl_context.load_verify_locations(capath=str(ca_path.absolute()))

    elif ca_path.exists():
        ssl_context.load_verify_locations(cafile=str(ca_path.absolute()))

    else:
        raise ValueError(f"Certificate CA verify path '{location}' does not exist")



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import dataclasses
import typing as t

from ._messages import Command, Request
from ._types import (
    ExceptionFilterOptions,
    ExceptionOptions,
    Source,
    SourceBreakpoint,
    StackFrameFormat,
    ValueFormat,
)


@dataclasses.dataclass()
class AttachRequest(Request, dap={"arguments": {"__restart": "restart"}}):
    """Attach DA to a debuggee that is already running.

    Sent from the client to the debug adapter to attach itself to a debuggee
    that is already running.

    Args:
        arguments: The arguments from the client, the structure of this is
            dependent on the client and what was requested.
        restart: Arbitrary data from the previous, restarted session.
    """

    command = Command.ATTACH

    arguments: t.Dict[str, t.Any]
    restart: t.Any = None

    def pack(self) -> dict[str, t.Any]:
        value = super().pack()
        value["arguments"].update(self.arguments)

        return value

    @classmethod
    def unpack(cls, data: dict[str, t.Any]) -> AttachRequest:
        arguments = data["arguments"]
        restart = arguments.pop("__restart", None)
        obj = AttachRequest(
            arguments=arguments,
            restart=restart,
        )
        object.__setattr__(obj, "seq", data["seq"])

        return obj


@dataclasses.dataclass()
class CancelRequest(Request, dap={"arguments": {"requestId": "request_id", "progressId": "progress_id"}}):
    """Cancel a request made by the client.

    Sent from the client to the debug adapter to cancel a request that it is
    no longer interested in or cancel a progress sequence.

    Args:
        request_id: The seq of the request to cancel. If no seq is specified
            then all are cancelled.
        progress_id: The seq of the progress to cancel.
    """

    command = Command.CANCEL

    request_id: t.Optional[int] = None
    progress_id: t.Optional[int] = None


@dataclasses.dataclass()
class ConfigurationDoneRequest(Request):
    """Client has finished initialization of the debug adapter.

    Sent by the client to signal it is done with the initial configuration of
    the debug adapter.
    """

    command = Command.CONFIGURATION_DONE


@dataclasses.dataclass()
class ContinueRequest(Request, dap={"arguments": {"threadId": "thread_id", "singleThread": "single_thread"}}):
    """Request execution to resume.

    Sent by the client to request the thread execution to resume.

    Args:
        thread_id: The thread to resume
        single_thread: Resume only the thread specified or all threads if
            False.
    """

    command = Command.CONTINUE

    thread_id: int
    single_thread: bool = True


@dataclasses.dataclass()
class DisconnectRequest(
    Request,
    dap={
        "arguments": {
            "restart": "restart",
            "terminateDebuggee": "terminate_debuggee",
            "suspendDebuggee": "suspend_debuggee",
        }
    },
):
    """Asks the DA to disconnect from the debuggee.

    Sent by the client to ask the DA to disconnect from the debuggee and
    end the debug session.

    Args:
        restart: This request is part of a restart sequence.
        terminate_debuggee: Indicates whether the debuggee should be terminated
            when the debugger is disconnected. If None the DA is free to do
            whatever it thinks is best.
        suspend_debuggee: Indicates whether the debuggee should stay suspended
            when the debugger is disconnected.
    """

    command = Command.DISCONNECT

    restart: bool = False
    terminate_debuggee: t.Optional[bool] = None
    suspend_debuggee: bool = False


@dataclasses.dataclass()
class EvaluateRequest(
    Request,
    dap={
        "arguments": {
            "expression": "expression",
            "frameId": "frame_id",
            "context": "context",
            "format": "format",
        }
    },
):
    """Evaluate given expression.

    Sent by the client to evaluate a given expression.

    Args:
        expression: THe expression to evaluate.
        frame_id: The expression in the scope of this stack frame. If not
            specified, the expression is evaluated in the global scope.
        context: The context in which the evaluate request is used.
        format: Details on how to format the result.
    """

    command = Command.EVALUATE

    expression: str
    frame_id: t.Optional[int] = None
    context: t.Optional[t.Union[str, t.Literal["variables", "watch", "repl", "hover", "clipboard"]]] = None
    format: t.Optional[ValueFormat] = None


@dataclasses.dataclass()
class InitializeRequest(
    Request,
    dap={
        "arguments": {
            "clientID": "client_id",
            "clientName": "client_name",
            "adapterID": "adapter_id",
            "locale": "locale",
            "linesStartAt1": "lines_start_at_1",
            "columnsStartAt1": "columns_start_at_1",
            "pathFormat": "path_format",
            "supportsVariableType": "supports_variable_type",
            "supportsVariablePaging": "supports_variable_paging",
            "supportsRunInTerminalRequest": "supports_run_in_terminal_request",
            "supportsMemoryReferences": "supports_memory_references",
            "supportsProgressReporting": "supports_progress_reporting",
            "supportsInvalidatedEvent": "supports_invalidated_event",
            "supportsMemoryEvent": "supports_memory_event",
        }
    },
):
    """Initialize DA with client capabilities.

    Send by the client to the debug adapter as the first message. It contains
    the capabilities of the client and expects the DA capabilities in the
    InitializeResponse.

    Args:
        adapter_id: The ID of the debug adapter.
        client_id: The ID of the client using this adapter.
        client_name: The name of the client using this adapter.
        locale: The ISO-639 locale of the client using this adapter.
        lines_start_at_1: All line numbers are 1-based, else 0-based.
        columns_start_at_1: All column numbers are 1-based, else 0-based.
        path_format: The format that paths are specified as.
        supports_variable_type: Client supports the optional type attribute for
            variables.
        supports_variable_paging: Client supports the paging of variables.
        supports_run_in_terminal_request: Client supports the
            RunInTerminalRequest.
        supports_memory_references: Client supports memory references.
        supports_progress_reporting: Client supports progress reporting.
        supports_invalidated_event: Client supports the invalidated event.
        supports_memory_event: Client supports the memory event.
    """

    command = Command.INITIALIZE

    adapter_id: str
    client_id: t.Optional[str] = None
    client_name: t.Optional[str] = None
    locale: t.Optional[str] = None
    lines_start_at_1: bool = True
    columns_start_at_1: bool = True
    path_format: t.Literal["path", "uri"] = "path"
    supports_variable_type: bool = False
    supports_variable_paging: bool = False
    supports_run_in_terminal_request: bool = False
    supports_memory_references: bool = False
    supports_progress_reporting: bool = False
    supports_invalidated_event: bool = False
    supports_memory_event: bool = False


@dataclasses.dataclass()
class LaunchRequest(Request, dap={"arguments": {"noDebug": "no_debug", "__restart": "restart"}}):
    """Request to start the debuggee with or without debugging.

    This is a request sent from the client to the debug adapter to start the
    debuggee with or without debugging.

    Args:
        arguments: The arguments from the client, the structure of this is
            dependent on the client and what was requested.
        no_debug: The program should launch without enabling debugging.
        restart: Optional data from the previous, restarted session.
    """

    command = Command.LAUNCH

    arguments: t.Dict[str, t.Any]
    no_debug: bool = False
    restart: t.Any = None

    def pack(self) -> dict[str, t.Any]:
        value = super().pack()
        value["arguments"].update(self.arguments)

        return value

    @classmethod
    def unpack(cls, data: dict[str, t.Any]) -> LaunchRequest:
        arguments = data["arguments"]
        no_debug = arguments.pop("noDebug", False)
        restart = arguments.pop("__restart", None)
        obj = LaunchRequest(
            arguments=arguments,
            no_debug=no_debug,
            restart=restart,
        )
        object.__setattr__(obj, "seq", data["seq"])

        return obj


@dataclasses.dataclass()
class NextRequest(
    Request,
    dap={
        "arguments": {
            "threadId": "thread_id",
            "singleThread": "single_thread",
            "granularity": "granularity",
        }
    },
):
    """Request to execute one step.

    This is a request sent from the client to the debug adapter to execute the
    next step and send a stop request.

    Args:
        thread_id: The thread for which to resume execution for one step.
        single_thread: All other suspended threads are not resumed.
        granularity: The granularity stepping.
    """

    command = Command.NEXT

    thread_id: int
    single_thread: bool = False
    granularity: t.Literal["statement", "line", "instruction"] = "statement"


@dataclasses.dataclass()
class RunInTerminalRequest(
    Request,
    dap={
        "arguments": {
            "kind": "kind",
            "title": "title",
            "cwd": "cwd",
            "args": "args",
            "env": "env",
        }
    },
):
    """Request from the debug adapter to run a command.

    This is a request from the debug adapter to the client to run a command in
    the client's terminal. This is typically used to launch the debuggee in a
    terminal provided by the client.

    Args:
        kind: The type of terminal to launch with.
        title: The title of the terminal.
        cwd: The working directory of the command.
        args: List of arguments, including the executable, to run the command
            with.
        env: Optional environment key-value pairs that are added to or removed
            from the default environment.
    """

    command = Command.RUN_IN_TERMINAL

    kind: t.Literal["integrated", "external"] = "integrated"
    title: t.Optional[str] = None
    cwd: str = ""
    args: t.List[str] = dataclasses.field(default_factory=list)
    env: t.Dict[str, t.Optional[str]] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class ScopesRequest(Request, dap={"arguments": {"frameId": "frame_id"}}):
    """Request variables scopes.

    Requests variable scopes for a given stackframe ID.

    Args:
        frame_id: The stackframe identifier to retrieve the scopes for.
    """

    command = Command.SCOPES

    frame_id: int


@dataclasses.dataclass
class SetBreakpointsRequest(
    Request,
    dap={
        "arguments": {
            "source": "source",
            "breakpoints": "breakpoints",
            "lines": "lines",
            "sourceModified": "source_modified",
        }
    },
):
    """Set breakpoint for a source.

    Used by the client to set breakpoints and clear any previous breakpoints
    for a source.

    Args:
        source: The source location of the breakpoints.
        breakpoints: The locations of the breakpoints in the source.
        lines: Deprecated but the lines in the source that are breakpoints.
        source_modified: Indicates the underlying source has been modified
            which results in new breakpoint locations.
    """

    command = Command.SET_BREAKPOINTS

    source: Source
    breakpoints: t.List[SourceBreakpoint] = dataclasses.field(default_factory=list)
    lines: t.List[int] = dataclasses.field(default_factory=list)
    source_modified: bool = False


@dataclasses.dataclass()
class SetExceptionBreakpointsRequest(
    Request,
    dap={
        "arguments": {
            "filters": "filters",
            "filterOptions": "filter_options",
            "exceptionOptions": "exception_options",
        }
    },
):
    """Configure exception breakpoint behavior.

    Sent by the client to configure the debugger's response to thrown
    exceptions.

    Args:
        filters: Set of exception filters by ID as specified by the
            Capabilities.exception_breakpoint_filters sent by the debug
            adapter.
        filter_options: List of exception filters and their options.
        exception_options: Configuration options for the selected exceptions.
    """

    command = Command.SET_EXCEPTION_BREAKPOINTS

    filters: t.List[str] = dataclasses.field(default_factory=list)
    filter_options: t.List[ExceptionFilterOptions] = dataclasses.field(default_factory=list)
    exception_options: t.List[ExceptionOptions] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class SetVariableRequest(
    Request,
    dap={
        "arguments": {
            "variablesReference": "variables_reference",
            "name": "name",
            "value": "value",
            "format": "format",
        }
    },
):
    """Set the variable with a new value.

    Sets a variable in a given context with a new value.

    Args:
        variables_reference: The variable container id.
        name: The name of the variable in the container.
        value: The value of the variable to set.
        format: Optional details on how the value is formatted.
    """

    command = Command.SET_VARIABLE

    variables_reference: int
    name: str
    value: str
    format: t.Optional[ValueFormat] = None


@dataclasses.dataclass()
class StackTraceRequest(
    Request,
    dap={
        "arguments": {
            "threadId": "thread_id",
            "startFrame": "start_frame",
            "levels": "levels",
            "format": "format",
        }
    },
):
    """Request a stacktrace.

    Sent by the client to request a stacktrace from the current execution
    state of a given thread.

    Args:
        thread_id: The stacktrace for this thread.
        start_frame: Index of the first frame to return.
        levels: The maximum number of frames to return.
        format: Details on how to format the stack frames.
    """

    command = Command.STACK_TRACE

    thread_id: int
    start_frame: t.Optional[int] = None
    levels: t.Optional[int] = None
    format: t.Optional[StackFrameFormat] = None


@dataclasses.dataclass()
class StepInRequest(
    Request,
    dap={
        "arguments": {
            "threadId": "thread_id",
            "singleThread": "single_thread",
            "targetId": "thread_id",
            "granularity": "granularity",
        }
    },
):
    """Request to step into a function/method.

    Sent by the client to request the given thread to step into a function
    and allows all other threads to run freely by resuming them.

    Args:
        thread_id: The thread for which to resume execution for on step into.
        single_thread: If True, all other suspended threads are not resumed.
        target_id: The id of the target to step into.
        granularity: Granularity level to step.
    """

    command = Command.STEP_IN

    thread_id: int
    single_thread: bool = False
    target_id: t.Optional[int] = None
    granularity: t.Literal["statement", "line", "instruction"] = "statement"


@dataclasses.dataclass()
class StepOutRequest(
    Request,
    dap={
        "arguments": {
            "threadId": "thread_id",
            "singleThread": "single_thread",
            "granularity": "granularity",
        }
    },
):
    """Request to ste out from a function/method.

    Sent by the client to request the given thread to step out of a function
    and allows all the other threads to run freely by resuming them.

    Args:
        thread_id: The thread for which to resume execution for one step out.
        single_thread: If True, all other suspended threads are not resumed.
        granularity: Granularity level to step.
    """

    command = Command.STEP_OUT

    thread_id: int
    single_thread: bool = False
    granularity: t.Literal["statement", "line", "instruction"] = "statement"


@dataclasses.dataclass()
class TerminateRequest(Request, dap={"arguments": {"restart": "restart"}}):
    """Request to gracefully terminate the debuggee.

    Request sent by the client to gracefully terminate the debuggee. This is
    sent before a DisconnectRequest as a second attempt if the terminate didn't
    occur.

    Args:
        restart: The request is part of a restart sequence.
    """

    command = Command.TERMINATE

    restart: bool = False


@dataclasses.dataclass()
class ThreadsRequest(Request):
    """Request to retrieve a list of all thread.

    Request sent by the client to retrieve all threads of the debuggee.
    """

    command = Command.THREADS


@dataclasses.dataclass()
class VariablesRequest(
    Request,
    dap={
        "arguments": {
            "variablesReference": "variables_reference",
            "filter": "filter",
            "start": "start",
            "count": "count",
            "format": "format",
        }
    },
):
    """Retrieves variables.

    Request sent by the client to retrieve all child variables for the given
    variable reference.

    Args:
        variables_reference: The reference id to retrieve.
        filter: Optionally filter child variables to either named or indexed
            variables.
        start: The index of the first variable to return.
        count: The number of variables to return or 0 for all.
        format: Details on how to format the variable values.
    """

    command = Command.VARIABLES

    variables_reference: int
    filter: t.Optional[t.Literal["indexed", "named"]] = None
    start: t.Optional[int] = None
    count: t.Optional[int] = None
    format: t.Optional[ValueFormat] = None



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import dataclasses
import enum
import typing as t

from ._messages import Event, EventType
from ._types import Breakpoint, Source


class StoppedReason(enum.Enum):
    STEP = "step"
    BREAKPOINT = "breakpoint"
    EXCEPTION = "exception"
    PAUSE = "pause"
    ENTRY = "entry"
    GOTO = "goto"
    FUNCTION_BREAKPOINT = "function breakpoint"
    DATA_BREAKPOINT = "data breakpoint"
    INSTRUCTION_BREAKPOINT = "instruction breakpoint"


@dataclasses.dataclass()
class BreakpointEvent(Event, dap={"body": {"reason": "reason", "breakpoint": "breakpoint"}}):
    """Indicate information about a breakpoint has changed.

    Event sent by the debuggee that indicates that some information about a
    breakpoint has changed.

    Args:
        reason: The reason for the breakpoint event.
        breakpoint: The id of this value is used to select the breakpoint on
            the client, the other values are used as the new information.
    """

    event = EventType.BREAKPOINT

    reason: t.Union[str, t.Literal["changed", "new", "removed"]]
    breakpoint: Breakpoint


@dataclasses.dataclass()
class ExitedEvent(Event, dap={"body": {"exitCode": "exit_code"}}):
    """Indicate the debuggee has existed.

    Event sent by the debuggee that indicates it has exited with the provided
    exit code.

    Args:
        exit_code: The exit code returned from the debuggee.
    """

    event = EventType.EXITED

    exit_code: int


@dataclasses.dataclass()
class InitializedEvent(Event):
    """Indicate the DA is ready to accept configuration requests.

    A DA is expected to send this event when it is ready to accept
    configuration requests.
    """

    event = EventType.INITIALIZED


@dataclasses.dataclass()
class OutputEvent(
    Event,
    dap={
        "body": {
            "category": "category",
            "output": "output",
            "group": "group",
            "variablesReference": "variables_reference",
            "source": "source",
            "line": "line",
            "column": "column",
            "data": "data",
        }
    },
):
    """Indicates that the target has produced some output.

    A debuggee is expected to send this event when it has produced some output
    for the client to process.

    Args:
        category: The output category.
        output: The output to report.
        group: The identifier for grouping related messages.
        variables_reference: Related variables.
        source: The source location where the output was produced.
        line: The source location's line where the output was produced.
        column: The source location's column where the output was produced.
        data: Additional data to report.
    """

    event = EventType.OUTPUT

    category: t.Union[t.Literal["console", "important", "stdout", "stderr", "telemetry"], str]
    output: str
    group: t.Optional[t.Literal["start", "startCollapsed", "end"]] = None
    variables_reference: t.Optional[int] = None
    source: t.Optional[Source] = None
    line: t.Optional[int] = None
    column: t.Optional[int] = None
    data: t.Any = None


@dataclasses.dataclass()
class StoppedEvent(
    Event,
    dap={
        "body": {
            "reason": "reason",
            "description": "description",
            "threadId": "thread_id",
            "preserveFocusHint": "preserve_focus_hint",
            "text": "text",
            "allThreadsStopped": "all_threads_stopped",
            "hitBreakpointIds": "hit_breakpoint_ids",
        },
    },
):
    """Indicates debuggee has stopped due to some condition.

    Event sent by the debuggee that indicates it has stopped due to some
    condition, like a breakpoint, or a stepping request is completed.

    Args:
        reason: The reason for the event.
        description: Full reason, displayed in the UI as is.
        thread_id: The thread which was stopped.
        preserve_focus_hint: The client should not change the focus.
        text: Additional information.
        all_threads_stopped: All threads are stopped and not just one.
        hit_breakpoint_ids: Ids of the breakpoints that triggers the event.
    """

    event = EventType.STOPPED

    reason: StoppedReason
    description: t.Optional[str] = None
    thread_id: t.Optional[int] = None
    preserve_focus_hint: bool = False
    text: t.Optional[str] = None
    all_threads_stopped: bool = False
    hit_breakpoint_ids: list[int] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class TerminatedEvent(Event, dap={"body": {"restart": "restart"}}):
    """Indicate the debuggee has terminated.

    Event sent by the DA that debugging of the debuggee has terminated. This
    does not mean the debuggee itself has exited.

    Args:
        restart: Opaque values not used by the client associated with this
            terminated event.
    """

    event = EventType.TERMINATED

    restart: t.Any = None


@dataclasses.dataclass()
class ThreadEvent(Event, dap={"body": {"reason": "reason", "threadId": "thread_id"}}):
    """Indicate a thread has started or exited.

    Event sent by the debuggee that indicates the status of a thread.

    Args:
        reason: The reason for the event, either started or exited.
        thread_id: The thread id the event is associated with.
    """

    event = EventType.THREAD

    reason: t.Literal["started", "exited"]
    thread_id: int



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

import typing as t


class Singleton(type):
    """Singleton used to ensure only instance of a class exists."""

    __instances: dict[type, object] = {}

    def __call__(
        cls,
        *args: t.Any,
        **kwargs: t.Any,
    ) -> object:
        if cls not in cls.__instances:
            cls.__instances[cls] = super().__call__(*args, **kwargs)

        return cls.__instances[cls]



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import dataclasses
import typing as t

from ._messages import Command, Response
from ._types import (
    Breakpoint,
    Capabilities,
    Message,
    Scope,
    StackFrame,
    Thread,
    Variable,
    VariablePresentationHint,
)


@dataclasses.dataclass()
class AttachResponse(Response):
    """Response to the AttachRequest."""

    command = Command.ATTACH


@dataclasses.dataclass()
class CancelResponse(Response):
    """Response to the CancelRequest."""

    command = Command.CANCEL


@dataclasses.dataclass()
class ConfigurationDoneResponse(Response):
    """Response to the ConfigurationDoneRequest."""

    command = Command.CONFIGURATION_DONE


@dataclasses.dataclass()
class ContinueResponse(Response, dap={"body": {"allThreadsContinued": "all_threads_continued"}}):
    """Response to the ContinueRequest."""

    command = Command.CONTINUE

    all_threads_continued: bool = True


@dataclasses.dataclass()
class DisconnectResponse(Response):
    """Response to the DisconnectRequest."""

    command = Command.DISCONNECT


@dataclasses.dataclass()
class EvaluateResponse(
    Response,
    dap={
        "body": {
            "result": "result",
            "type": "type",
            "presentationHint": "presentation_hint",
            "variablesReference": "variables_reference",
            "namedVariables": "named_variables",
            "indexedVariables": "indexed_variables",
            "memoryReference": "memory_reference",
        }
    },
):
    """Response to the EvaluateRequest."""

    command = Command.EVALUATE

    result: str
    type: t.Optional[str] = None
    presentation_hint: t.Optional[VariablePresentationHint] = None
    variables_reference: int = 0
    named_variables: t.Optional[int] = None
    indexed_variables: t.Optional[int] = None
    memory_reference: t.Optional[str] = None


@dataclasses.dataclass()
class InitializeResponse(Response, dap={"body": "capabilities"}):
    """Response to InitializeRequest

    The response sent to the client of the InitializeRequest message.

    Args:
        capabilities: The DA capabilities.
    """

    command = Command.INITIALIZE

    capabilities: Capabilities


@dataclasses.dataclass()
class LaunchResponse(Response):
    """Response to a LaunchRequest."""

    command = Command.LAUNCH


@dataclasses.dataclass()
class NextResponse(Response):
    """Response to a NextRequest."""

    command = Command.NEXT


@dataclasses.dataclass()
class RunInTerminalResponse(Response, dap={"body": {"processId": "process_id", "shellProcessId": "shell_process_id"}}):
    """Response to RunInTerminalRequest.

    The response to a RunInTerminalRequest. It is expected either process_id or
    shell_process_id is set, not both.

    Args:
        process_id: The process id
        shell_process_id: The process id of the terminal shell.
    """

    command = Command.RUN_IN_TERMINAL

    process_id: t.Optional[int] = None
    shell_process_id: t.Optional[int] = None


@dataclasses.dataclass()
class ScopesResponse(Response, dap={"body": {"scopes": "scopes"}}):
    """Response to ScopesRequest.

    The response to ScopesRequest that requests the variable scopes for a given
    stackframe ID.

    Args:
        scopes: The variable scopes.
    """

    command = Command.SCOPES

    scopes: t.List[Scope] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class SetBreakpointsResponse(Response, dap={"body": {"breakpoints": "breakpoints"}}):
    """Response to SetBreakpointsRequest.

    The response to SetBreakpointRequest that returns the information about
    all the breakpoints that were or were not set.

    Args:
        breakpoints: The breakpoints in the same order of the request and their
            status.
    """

    command = Command.SET_BREAKPOINTS

    breakpoints: t.List[Breakpoint] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class SetExceptionBreakpointsResponse(Response, dap={"body": {"breakpoints": "breakpoints"}}):
    """Response to SetExceptionBreakpointsRequest.

    This is the response to the SetExceptionBreakpointsRequest that contains
    the breakpoint information requested in the request. The breakpoints must
    be in the same order as the requested order.

    Args:
        breakpoints: The breakpoint information that was requested.
    """

    command = Command.SET_EXCEPTION_BREAKPOINTS

    breakpoints: t.List[Breakpoint] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class SetVariableResponse(
    Response,
    dap={
        "body": {
            "value": "value",
            "type": "type",
            "variablesReference": "variables_reference",
            "namedVariables": "named_variables",
            "indexedVariables": "indexed_variables",
        }
    },
):
    """Response to SetVariableRequest."""

    command = Command.SET_VARIABLE

    value: str
    type: t.Optional[str] = None
    variables_reference: t.Optional[int] = None
    named_variables: t.Optional[int] = None
    indexed_variables: t.Optional[int] = None


@dataclasses.dataclass()
class StackTraceResponse(Response, dap={"body": {"stackFrames": "stack_frames", "totalFrames": "total_frames"}}):
    """Response to a StackTraceRequest.

    A response to a StackTraceRequest that returns the stack frames for the
    thread(s) requested.

    Args:
        stack_frames: The frames of the stack frame.
        total_frames: The total number of frames available in the stack.
    """

    command = Command.STACK_TRACE

    stack_frames: t.List[StackFrame] = dataclasses.field(default_factory=list)
    total_frames: t.Optional[int] = None


@dataclasses.dataclass()
class StepInResponse(Response):
    """Response to a StepInRequest."""

    command = Command.STEP_IN


@dataclasses.dataclass()
class StepOutResponse(Response):
    """Response to a StepOutRequest."""

    command = Command.STEP_OUT


@dataclasses.dataclass()
class TerminateResponse(Response):
    """Response to a TerminateRequest."""

    command = Command.TERMINATE


@dataclasses.dataclass()
class ThreadsResponse(Response, dap={"body": {"threads": "threads"}}):
    """Response to a ThreadsRequest.

    A response to a ThreadRequest that is sent to the client.

    Args:
        threads: The threads on the debuggee.
    """

    command = Command.THREADS

    threads: t.List[Thread] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class VariablesResponse(Response, dap={"body": {"variables": "variables"}}):
    """Response to a VariablesRequest.

    A response to a VariablesRequest that is sent to the client.

    Args:
        variables: The variables.
    """

    command = Command.VARIABLES

    variables: t.List[Variable] = dataclasses.field(default_factory=list)


@dataclasses.dataclass()
class ErrorResponse(Response, dap={"message": "message", "body": {"error": "error"}}):
    """Error response to client.

    Sent by the debug adapter to the client to provide more error details.

    Args:
        request_seq: The seq of the request this is responding to.
        command: The command of the request this is an error for.
        message: Error message in short form.
        error: Optional structured error message.
    """

    success: bool = dataclasses.field(init=False, default=False)
    command: Command = dataclasses.field(init=True)
    message: t.Optional[str] = None
    error: t.Optional[Message] = None



# Copyright (c) 2023 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import logging
import typing as t

LogLevel = t.Literal["info", "debug", "warning", "error"]

DEFAULT_FORMAT = "%(asctime)s | %(name)s | %(filename)s:%(lineno)s %(funcName)s() %(message)s"


def configure_file_logging(
    file: str,
    level: LogLevel,
    format: str | None = None,
) -> None:
    """Configures the ansibug logger for file logging.

    Configures the file logging config for the ansibug logger.

    Args:
        file: The log file to log to.
        level: The log level to set.
        format: A custom format to use for the log messages.
    """
    log_level = {
        "info": logging.INFO,
        "debug": logging.DEBUG,
        "warning": logging.WARNING,
        "error": logging.ERROR,
    }[level]

    fh = logging.FileHandler(file, mode="a", encoding="utf-8")
    fh.setLevel(log_level)
    fh.setFormatter(logging.Formatter(format or DEFAULT_FORMAT))

    ansibug_logger = logging.getLogger("ansibug")
    ansibug_logger.setLevel(log_level)
    ansibug_logger.addHandler(fh)



# Copyright (c) 2022 Jordan Borean
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import dataclasses
import enum
import json
import queue
import sys
import typing as t


class MessageType(enum.Enum):
    REQUEST = "request"
    RESPONSE = "response"
    EVENT = "event"


class Command(enum.Enum):
    ATTACH = "attach"
    CANCEL = "cancel"
    CONFIGURATION_DONE = "configurationDone"
    CONTINUE = "continue"
    DISCONNECT = "disconnect"
    EVALUATE = "evaluate"
    INITIALIZE = "initialize"
    LAUNCH = "launch"
    NEXT = "next"
    RUN_IN_TERMINAL = "runInTerminal"
    SET_BREAKPOINTS = "setBreakpoints"
    SET_EXCEPTION_BREAKPOINTS = "setExceptionBreakpoints"
    SET_VARIABLE = "setVariable"
    STACK_TRACE = "stackTrace"
    STEP_IN = "stepIn"
    STEP_OUT = "stepOut"
    TERMINATE = "terminate"
    THREADS = "threads"
    SCOPES = "scopes"
    VARIABLES = "variables"


class EventType(enum.Enum):
    INITIALIZED = "initialized"
    STOPPED = "stopped"
    CONTINUED = "continued"
    EXITED = "exited"
    TERMINATED = "terminated"
    THREAD = "thread"
    OUTPUT = "output"
    BREAKPOINT = "breakpoint"
    MODULE = "module"
    LOADED_SOURCE = "loadedSource"
    PROCESS = "process"
    CAPABILITIES = "capabilities"
    PROGRESS_START = "progressStart"
    PROGRESS_UPDATE = "progressUpdate"
    PROGRESS_END = "progressEnd"
    INVALIDATED = "invalidated"
    MEMORY = "memory"


class DAPEncoder(json.JSONEncoder):
    def default(self, obj: t.Any) -> t.Any:
        if isinstance(obj.__class__, DAPObjectMeta) and hasattr(obj, "pack"):
            return obj.pack()

        elif isinstance(obj, enum.Enum):
            return obj.value

        else:
            return str(obj)


def _parse_field_type(obj: type, value: str) -> t.Callable[[t.Any], t.Any] | None:
    base_globals = getattr(sys.modules.get(obj.__module__, None), "__dict__", {})
    base_locals = dict(vars(obj))
    evaled_type = eval(value, base_globals, base_locals)

    attr_type = None
    if isinstance(evaled_type, type):
        attr_type = evaled_type
    else:
        type_origin = t.get_origin(evaled_type)
        if type_origin is t.Union:
            union_types = t.get_args(evaled_type)
            if len(union_types) == 2 and type(None) in union_types:
                attr_type = union_types[0]
        elif type_origin is list:
            attr_type = t.get_args(evaled_type)[0]

    if not attr_type:
        return None

    elif isinstance(attr_type, type) and issubclass(attr_type, enum.Enum):
        return attr_type

    elif isinstance(attr_type, DAPObjectMeta) and (unpack := getattr(attr_type, "unpack")):
        return unpack

    else:
        return None


class DAPObjectMeta(type):
    EVENT_TYPE: type | None = None
    REQUEST_TYPE: type | None = None
    RESPONSE_TYPE: type | None = None
    REGISTRY: dict[str, dict[str, t.Callable[[dict[str, t.Any]], ProtocolMessage]]] = {
        MessageType.EVENT.value: {},
        MessageType.REQUEST.value: {},
        MessageType.RESPONSE.value: {},
    }

    def __new__(
        mcs: type[DAPObjectMeta],
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, t.Any],
        dap: dict[str, t.Any] | None = None,
    ) -> DAPObjectMeta:
        cls = super().__new__(mcs, name, bases, namespace)

        # Build the DAP key mapping dictionary for each type, merge it with the
        # base class definition.
        dap = dap or {}
        for base in reversed(bases):
            dap = getattr(base, "_dap", {}) | dap

        # We cannot refer to these types directly as they are created by this
        # metaclass, store them for easier referencing later.
        if name == "Event" and namespace["__module__"] == DAPObjectMeta.__module__:
            DAPObjectMeta.EVENT_TYPE = cls

        elif name == "Request" and namespace["__module__"] == DAPObjectMeta.__module__:
            DAPObjectMeta.REQUEST_TYPE = cls

        elif name == "Response" and namespace["__module__"] == DAPObjectMeta.__module__:
            DAPObjectMeta.RESPONSE_TYPE = cls

        setattr(cls, "_dap", dap)

        # Only add pack to the first class in inheritance.
        if "pack" not in cls.__dict__:
            setattr(cls, "pack", DAPObjectMeta.default_pack)

        # Use the explicitly defined unpack class method if it's present
        # otherwise use ours.
        if not (unpack := namespace.get("unpack", None)):
            unpack = lambda d: DAPObjectMeta.default_unpack(cls, d)
            setattr(cls, "unpack", unpack)
        else:
            unpack = getattr(cls, "unpack")

        # Register the unpack methods for each message according to the type
        # and type identifier.
        if (
            DAPObjectMeta.EVENT_TYPE
            and issubclass(cls, DAPObjectMeta.EVENT_TYPE)
            and cls != DAPObjectMeta.EVENT_TYPE
            and (event := namespace.get("event", None))
            and isinstance(event, EventType)
        ):
            DAPObjectMeta.REGISTRY[MessageType.EVENT.value][event.value] = unpack

        elif (
            DAPObjectMeta.REQUEST_TYPE
            and issubclass(cls, DAPObjectMeta.REQUEST_TYPE)
            and cls != DAPObjectMeta.REQUEST_TYPE
            and (command := namespace.get("command", None))
            and isinstance(command, Command)
        ):
            DAPObjectMeta.REGISTRY[MessageType.REQUEST.value][command.value] = unpack

        elif (
            DAPObjectMeta.RESPONSE_TYPE
            and issubclass(cls, DAPObjectMeta.RESPONSE_TYPE)
            and cls != DAPObjectMeta.RESPONSE_TYPE
        ):
            if (command := namespace.get("command", None)) and isinstance(command, Command):
                DAPObjectMeta.REGISTRY[MessageType.RESPONSE.value][command.value] = unpack
            else:
                # ErrorRecord is special where it's not a specific command but
                # is handled in a special way during the unpacking phase.
                DAPObjectMeta.REGISTRY[MessageType.RESPONSE.value]["error"] = unpack

        return cls

    def default_pack(
        self: t.Any,
    ) -> dict[str, t.Any]:
        packed_value: dict[str, t.Any] = {}

        todo: queue.Queue[tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]] = queue.Queue()
        todo.put((getattr(self.__class__, "_dap", {}), packed_value))

        while True:
            try:
                mapping, value = todo.get(block=False)
            except queue.Empty:
                break

            for key, info in mapping.items():
                if key == "__types":
                    continue
                elif isinstance(info, dict):
                    todo.put((info, value.setdefault(key, {})))
                else:
                    value[key] = getattr(self, info)

        return packed_value

    def default_unpack(
        cls,
        data: dict[str, t.Any],
    ) -> t.Any:
        dap = getattr(cls, "_dap", {})
        type_info = dap.get("__types", {})
        if not type_info and dataclasses.is_dataclass(cls):
            field_info = dataclasses.fields(cls)  # type: ignore[unreachable]

            type_info = {}
            cls_fields = {}
            for base_cls in cls.__mro__:
                if not dataclasses.is_dataclass(base_cls):
                    continue

                cls_fields.update({k: base_cls for k in base_cls.__dataclass_fields__.keys()})

            for field in field_info:
                field_unpacker = None
                base_cls = cls_fields[field.name]
                field_unpacker = _parse_field_type(base_cls, str(field.type))
                type_info[field.name] = (field.init, field_unpacker)

            dap["__types"] = type_info

        kwargs: dict[str, t.Any] = {}
        manual_set: dict[str, t.Any] = {}

        todo: queue.Queue[tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]] = queue.Queue()
        todo.put((dap, data))

        while True:
            try:
                mapping, body = todo.get(block=False)
            except queue.Empty:
                break

            for key, value in body.items():
                if key not in mapping or value is None:
                    continue

                field_meta = mapping[key]
                if isinstance(field_meta, dict):
                    todo.put((field_meta, value))
                    continue

                can_init, unpack_func = type_info.get(field_meta, None)
                if not unpack_func:
                    unpack_func = lambda v: v

                dict_to_update = kwargs if can_init else manual_set
                if isinstance(value, list):
                    dict_to_update[field_meta] = [unpack_func(v) for v in value]
                else:
                    dict_to_update[field_meta] = unpack_func(value)

        new_obj = cls(**kwargs)
        for key, value in manual_set.items():
            object.__setattr__(new_obj, key, value)

        return new_obj


@dataclasses.dataclass()
class ProtocolMessage(metaclass=DAPObjectMeta, dap={"seq": "seq", "type": "message_type"}):
    """Base class for all DAP messages."""

    seq: int = dataclasses.field(init=False, default=0)
    message_type: MessageType = dataclasses.field(init=False)

    def pack(self) -> dict[str, t.Any]:  # type: ignore[empty-body] # Defined in metaclass
        ...  # pragma: no cover

    @classmethod
    def unpack(
        cls,
        data: dict[str, t.Any],
    ) -> ProtocolMessage:
        """Unpack DAP JSON data.

        Unpacks the DAP JSON data object into a structured object. If the type
        or the type's identifier is unknown then a ValueError is raised.

        Args:
            data: The JSON data dictionary to unpack.

        Returns:
            ProtocolMessage: The unpacked message.
        """
        obj_type = data["type"]

        registry = DAPObjectMeta.REGISTRY.get(obj_type, None)
        if not registry:
            raise ValueError(f"Unknown DAP message type '{obj_type}'")

        if obj_type == MessageType.REQUEST.value:
            cmd = data["command"]

            request_unpack = registry.get(cmd, None)
            if not request_unpack:
                raise ValueError(f"Unknown DAP request command '{cmd}'")

            return request_unpack(data)

        elif obj_type == MessageType.RESPONSE.value:
            cmd = data["command"]
            if not data["success"]:
                cmd = "error"

            response_unpack = registry.get(cmd, None)
            if not response_unpack:
                raise ValueError(f"Unknown DAP response command '{cmd}'")

            return response_unpack(data)

        else:
            event = data["event"]

            event_unpack = registry.get(event, None)
            if not event_unpack:
                raise ValueError(f"Unknown DAP event type '{event}'")

            return event_unpack(data)


@dataclasses.dataclass()
class Request(ProtocolMessage, dap={"command": "command"}):
    """Base class for DAP request messages."""

    message_type = MessageType.REQUEST

    command: Command = dataclasses.field(init=False)


@dataclasses.dataclass()
class Event(ProtocolMessage, dap={"event": "event"}):
    """Base class for DAP event messages."""

    message_type = MessageType.EVENT

    event: EventType = dataclasses.field(init=False)


@dataclasses.dataclass()
class Response(ProtocolMessage, dap={"request_seq": "request_seq", "success": "success", "command": "command"}):
    """Base class for DAP response messages."""

    message_type = MessageType.RESPONSE

    success: bool = dataclasses.field(init=False, default=True)
    command: Command = dataclasses.field(init=False)
    request_seq: int
